
import json
from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from dataverse.etl import register_etl
from dataverse.utils.format import get_uuidv1


@register_etl
def data_ingestion___cultura_x___raw2ufl(spark, ufl, *args, **kwargs):
    """
    convert raw format to ufl with custom template
    """
    def templatev1(row):
        new_row = {}
        new_row['id'] = get_uuidv1()
        new_row['name'] = 'cultura_x'
        new_row['text'] = row['text']
        new_row['meta'] = json.dumps({
            'url': row['url'],
            'timestamp': row['timestamp'],
            'source': row['source'],
        })
        return new_row

    ufl = ufl.map(lambda x: templatev1(x))

    return ufl






"""
Load CSV data
"""

from dataverse.etl import register_etl
from dataverse.utils.format import huggingface2parquet
from dataverse.utils.format import load_huggingface_dataset
from typing import Union, List


@register_etl
def data_ingestion___csv___csv2raw(
    spark,
    path: Union[str, List[str]],
    repartition=20,
    verbose=True,
    *args,
    **kwargs
):
    """
    Args:
        spark (SparkSession): spark session
        path (str or list): csv path
        repartition (int): the number of partitions
        verbose (bool): whether to print the information of the dataset
    """
    if isinstance(path, str):
        path = [path]

    df = spark.read.csv(*path, header=True)
    rdd = df.rdd.repartition(repartition)
    rdd = rdd.map(lambda row: row.asDict())

    return rdd




# AWS
from .aws import AWSClient 
from .aws import EMRManager

from .aws import aws_check_credentials
from .aws import aws_get_state
from .aws import aws_set_state

# EC2
from .aws import aws_ec2_instance_at_az
from .aws import aws_ec2_instance_info
from .aws import aws_ec2_all_instance_info
from .aws import aws_ec2_get_price

# SSM
from .aws import aws_ssm_run_commands

# VPC
from .aws import aws_vpc_create
from .aws import aws_vpc_delete
from .aws import aws_subnet_create
from .aws import aws_subnet_delete
from .aws import aws_subnet_az
from .aws import aws_emr_security_group_create
from .aws import aws_security_group_delete
from .aws import aws_security_group_remove_dependency
from .aws import aws_gateway_create
from .aws import aws_gateway_delete
from .aws import aws_route_table_create
from .aws import aws_route_table_delete
from .aws import aws_route_table_asscociate_subnet

from .aws import aws_elastic_ip_allocate
from .aws import aws_elastic_ip_release
from .aws import aws_nat_gateway_create
from .aws import aws_nat_gateway_delete

from .aws import aws_iam_role_create
from .aws import aws_iam_role_delete
from .aws import aws_iam_instance_profile_create
from .aws import aws_iam_instance_profile_delete

# S3
from .aws import aws_s3_path_parse
from .aws import aws_s3_create_bucket
from .aws import aws_s3_delete_bucket
from .aws import aws_s3_read
from .aws import aws_s3_download
from .aws import aws_s3_upload
from .aws import aws_s3_write
from .aws import aws_s3_delete
from .aws import aws_s3_list_buckets
from .aws import aws_s3_ls 
from .aws import aws_s3_get_object_type



from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from dataverse.etl import register_etl
from dataverse.utils.format import huggingface2parquet
from dataverse.utils.format import load_huggingface_dataset
from dataverse.utils.format import get_uuidv1

from typing import Union, List
from omegaconf import ListConfig

import datasets


"""
Supported datasets:
https://huggingface.co/datasets/togethercomputer/RedPajama-Data-1T
https://huggingface.co/datasets/togethercomputer/RedPajama-Data-1T-Sample
"""



"""
1 stage data ingestion - default
====================================
direct loading ufl with one ETL process
"""
def convert2ufl(row):
    row['id'] = get_uuidv1()
    row['name'] = 'red_pajama'
    return row

@register_etl
def data_ingestion___red_pajama___parquet2ufl(spark, input_paths, repartition=20, *args, **kwargs):
    """
    convert parquet file to ufl
    """
    df = spark.read.parquet(*input_paths)
    rdd = df.rdd.repartition(repartition)
    rdd = rdd.map(lambda row: row.asDict())
    rdd = rdd.map(lambda x: convert2ufl(x))

    return rdd

@register_etl
def data_ingestion___red_pajama___hf2ufl(
    spark,
    name_or_path : Union[str, List[str]] = 'togethercomputer/RedPajama-Data-1T-Sample',
    split=None,
    from_disk=False,
    repartition=20,
    verbose=True,
    *args,
    **kwargs
):
    """
    convert huggingface dataset to ufl

    Args:
        spark (SparkSession): spark session
        name_or_path (str or list): the name or path of the huggingface dataset
        split (str): the split of the dataset
        from_disk (bool): whether to load from disk
            - no split is allowed when from_disk is True
        repartition (int): the number of partitions
        verbose (bool): whether to print the information of the dataset
    """
    dataset = load_huggingface_dataset(name_or_path, split=split, from_disk=from_disk)
    parquet_path = huggingface2parquet(dataset, verbose=verbose)

    df = spark.read.parquet(parquet_path)
    rdd = df.rdd.repartition(repartition)
    rdd = rdd.map(lambda row: row.asDict())
    rdd = rdd.map(lambda x: convert2ufl(x))

    return rdd


"""
2 stage data ingestion - default
====================================
loading ufl with custom template with two ETL process
"""
@register_etl
def data_ingestion___red_pajama___hf2raw(
    spark,
    name_or_path : Union[str, List[str]] = 'togethercomputer/RedPajama-Data-1T-Sample',
    split=None,
    repartition=20,
    verbose=True,
    *args,
    **kwargs
):
    """
    convert huggingface dataset to raw format as dict

    Args:
        spark (SparkSession): spark session
        name_or_path (str or list): the name or path of the huggingface dataset
        split (str): the split of the dataset
        repartition (int): the number of partitions
        verbose (bool): whether to print the information of the dataset
    """
    dataset = load_huggingface_dataset(name_or_path, split=split)
    parquet_path = huggingface2parquet(dataset, verbose=verbose)
    df = spark.read.parquet(parquet_path)
    rdd = df.rdd.repartition(repartition)
    rdd = rdd.map(lambda row: row.asDict())

    return rdd


@register_etl
def data_ingestion___red_pajama___raw2ufl_templatev1(spark, ufl, *args, **kwargs):
    """
    convert raw format to ufl with custom template
    """
    def templatev1(row):
        row['id'] = get_uuidv1()
        row['name'] = 'red_pajama'
        return row

    ufl = ufl.map(lambda x: templatev1(x))

    return ufl

@register_etl
def data_ingestion___red_pajama___raw2ufl_templatev2(spark, ufl, *args, **kwargs):
    ...
    return ufl




"""
language filtering from Common Crawl

This is a migration of the common crawl code to Dataverse.
some part of code is from facebookresearch/cc_net
https://github.com/facebookresearch/cc_net/blob/main/cc_net/split_by_lang.py
"""

from pyspark.rdd import RDD
from pyspark.sql import DataFrame


import os
import multiprocessing
import functools
import requests
import fasttext
from fasttext.FastText import _FastText
from typing import Union
from pathlib import Path

from dataverse.etl.registry import register_etl
from dataverse.utils.setting import SystemSetting


def load_fasttext(
    url = "https://dl.fbaipublicfiles.com/fasttext/supervised-models/lid.176.bin",
):
    """
    There is 2 issues found here
    - due to unserilizable fasttext problem, we need to load the model for every task
        - this is a problem, extremely slow
        - we need to load the model once and use it for all tasks
    - since this could lead to duplicated download, we need to check if the model is already downloaded
        - so far found no duplicated download, but if there is, hope to be fixed in the future
    """
    # FIXME: this is a manual check for duplicate download
    # rd_n = np.random.randint(0, 1000000)
    # print(rd_n, 'entered load_fasttext model!')

    # Get the lid.bin file for Fasttext
    cache_dir = SystemSetting().CACHE_DIR
    cache_dir = Path(f"{cache_dir}/.cache/dataverse/model")
    fasttext_path = cache_dir / "fasttext" / "bin" / "lid.bin"
    fasttext_path.parent.mkdir(parents=True, exist_ok=True) # Make directories if not existed

    if not fasttext_path.exists():
        # FIXME: this is a manual check for duplicate download
        # print(rd_n, 'downloading fasttext model!')
        response = requests.get(url, stream=True)

        # Raise exception if downloading is not successful
        response.raise_for_status() 
        with open(fasttext_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
    
    # FIXME: this is to suppress the warning message
    # return fasttext.load_model(str(fasttext_path))
    return _FastText(model_path=str(fasttext_path))

def language_predict_fasttext(row, model, top_k: int = 1, score_rounding: int = 2):
    text = row['text'].replace("\n", "")
    labels, scores = model.predict(text, k=top_k)
    labels = [l.replace("__label__", "") for l in labels]

    row['labels'] = labels
    row['scores'] = scores.round(score_rounding)

    return row

def language_predict_fasttext_by_partition(rows, top_k: int = 1, score_rounding: int = 2):
    # loaded for every partition
    model = load_fasttext()

    # FIXME: not possible to use multiprocessing here because of the model is not serializable
    # pool = multiprocessing.Pool(processes = os.cpu_count() or 0)
    # results = pool.imap(
    #     functools.partial(language_predict_fasttext, model=model, top_k=top_k),
    #     rows,
    # )
    for row in rows:
        yield language_predict_fasttext(row, model, top_k=top_k)


@register_etl
def quality___language___fasttext_filter(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    top_k=1,
    score_rounding=2,
    threshold=0.0,
    whitelist=None,
    blacklist=None,
    *args,
    **kwargs
):
    """
    language filtering using fasttext

    - if language score is below threshold, it will be removed

    - rules for `whitelist` and `blacklist`:
        - if `whitelist` and `blacklist` are both None, all languages will be kept
            - default

        - if `whitelist` is not None, only languages in the `whitelist` will be kept
        - if `blacklist` is not None, languages in the `blacklist` will be removed
        - if `whitelist` and `blacklist` are both not None, error will be raised

    args:
        subset (str): subset or columns to consider filter 
        top_k (int): number of top languages to keep after classification
            - if fasttext classified 3 languages, top_k=1 will keep the top language
                - [en, fr, de] -> [en]
            - if fasttext classified 3 languages, top_k=2 will keep the top 2 languages
                - [en, fr, de] -> [en, fr]
        score_rounding (int): number of decimal places to round the scores
        threshold (float): minimum score to keep the language
        whitelist (list): list of languages to keep
        blacklist (list): list of languages to remove
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    # detect language using fasttext
    data = data.mapPartitions(functools.partial(
        language_predict_fasttext_by_partition,
        top_k=top_k,
        score_rounding=score_rounding,
    ))

    # filter by threshold
    data = data.filter(lambda x: any(s >= threshold for s in x['scores'][:top_k]))

    # filter by whitelist and blacklist
    if whitelist is not None and blacklist is not None:
        raise ValueError("whitelist and blacklist cannot be both not None")
    elif whitelist is not None:
        data = data.filter(lambda x: any(l in whitelist for l in x['labels'][:top_k]))
    elif blacklist is not None:
        data = data.filter(lambda x: all(l not in blacklist for l in x['labels'][:top_k]))
    else:
        # otherwise, keep all languages
        ...

    # remove labels and scores
    data = data.map(lambda x: {k: v for k, v in x.items() if k != 'labels' and k != 'scores'})

    return data



"""
Interface for user setting
"""

import os
import json
from pathlib import Path

from dataverse.utils.setting.system import SystemSetting


class UserSetting:
    """
    Proxy for user setting CRUD, synchronized with the user_setting.json file

    To manage user API keys, passwords, or other sensitive information,
    you can use directly store the information in the user_setting.json file
    or use this class to store the information as proxy. 

    Also, this class is a singleton class, so you can use it anywhere in the code

    caveat:
        - this is just a storage and does not include any logic more than CRUD
        - anything more than CRUD is a responsibility of outside of this class

    what does it means resposibility of outside of this class:
        When user wants to save API key and if does not exist
        asking user to input the API key with stdin might be a good idea.
        But this is not the responsibility of this class.

        This class will only return API key exists or not and base on that,
        outside of this class will ask user to input the API key
        or raise error or whatever it needs to do.
    """
    # Singleton
    _initialized = False

    # TODO: system setting per user [Candidate]
    ...

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(UserSetting, cls).__new__(cls)
        return cls.instance

    def __init__(self):
        # when the class is initialized, this is called everytime
        # regardless of the singleton. So adding the flag to check
        if self._initialized:
            return

        # create the user setting path
        os.makedirs(f"{SystemSetting().CACHE_DIR}/.cache/dataverse/setting", exist_ok=True)
        self.user_setting_path = os.path.join(
            SystemSetting().CACHE_DIR,
            ".cache/dataverse/setting/user_setting.json"
        )

        # load the user setting, if not exist, empty dict will be assigned
        self.user_setting = self.load(self.user_setting_path)
        self._initialized = True

    def reset(self):
        """
        reset the setting
        """
        self.user_setting = {}
        self.sync_file()

    def sync_file(self):
        """
        sync (class -> file)
        """
        with open(self.user_setting_path, "w") as f:
            json.dump(self.user_setting, f, indent=4)

    def sync_class(self):
        """
        sync (file -> class)
        """
        # sync the file to make sure the dict is up-to-date
        self.user_setting = self.load(self.user_setting_path)

    def load(self, path):
        """
        Load the user setting file
        """
        # check if user setting file exists
        if not os.path.exists(path):
            return {}

        # read the file
        with open(path, "r") as f:
            json_file = json.load(f)

        return json_file

    def check_naming_convention(self, key):
        """
        1. only CAPITALIZED format
            - e.g. GITHUB_API (O)
            - e.g. github_api (X)
        2. only alphanumeric and underscore
            - e.g. GITHUB_API2 (O)
            - e.g. github-api (X)
            - e.g. GITHUB_@API (X)
        3. only one underscore between words
            - e.g. GITHUB__API (X)
        4. no underscore at the start/end of the key
            - e.g. _GITHUB_API (X)
            - e.g. GITHUB_API_ (X)
        """
        # 1. only CAPITALIZED format
        if key != key.upper():
            raise ValueError(f"key [ {key} ] is not in Capitalized format")

        # 2. only alphanumeric and underscore
        for char in key:
            if not char.isalnum() and char != "_":
                raise ValueError(f"key [ {key} ] should only contains alphanumeric and underscore")

        # 3. only one underscore between words
        if "_" in key:
            # check if there is only one underscore
            divided_keys = key.split("_")
            if "" in divided_keys:
                raise ValueError(f"key [ {key} ] contains more than one underscore")

        # 4. no underscore at the start/end of the key
        if key.startswith("_") or key.endswith("_"):
            raise ValueError(f"key [ {key} ] contains underscore at the start/end of the key")


    def get(self, key):
        """
        """
        self.sync_class()
        if key not in self.user_setting:
            raise KeyError(f"key [ {key} ] does not exist in USER setting")
        return self.user_setting[key]

    def set(self, key, value):
        """
        """
        self.check_naming_convention(key)
        self.user_setting[key] = value
        self.sync_file()

    # Support dot-like access, e.g. setting.GITHUB_API
    def __getattr__(self, key):
        if key in [
            "_initialized",
            "user_setting",
            "user_setting_path"
        ]:
            return super().__getattr__(key)
        else:
            return self.get(key)

    def __setattr__(self, key, value):
        if key in [
            "_initialized",
            "user_setting",
            "user_setting_path"
        ]:
            super().__setattr__(key, value)
        else:
            self.set(key, value)

    # Support dict-like access, e.g. setting["GITHUB_API"]
    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        self.set(key, value)

    def delete(self, key):
        """
        """
        if key in self.user_setting:
            self.user_setting.pop(key, None)
            self.sync_file()
        else:
            raise KeyError(f"key [ {key} ] does not exist in USER setting")

    def list(self):
        """
        List all settings
        """
        self.sync_class()
        print(self.user_setting)

    def __repr__(self):
        self.sync_class()
        return json.dumps(self.user_setting, indent=4)

    def __str__(self):
        self.sync_class()
        return json.dumps(self.user_setting, indent=4)




"""
API to use AWS EMR with spark-submit
"""

import os
import argparse
import importlib

from dataverse.etl import ETLPipeline

def import_dynamic_etls():
    """
    Import dynamic etls which was created by user.
    """
    dynamic_etl_path = "/home/hadoop/dataverse/dynamic_etl"
    try:
        files = os.listdir(dynamic_etl_path)
    except FileNotFoundError:
        return
    except Exception as e:
        raise e

    # Filter out non-Python files
    files = [f for f in files if f.endswith('.py')]

    # Dynamically import all Python files in the directory
    for file in files:
        file_path = os.path.join(dynamic_etl_path, file)

        # Remove .py at the end
        module_name = file[:-3]

        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)


def main(config, verbose=False):
    """Main entry point for the aws emr."""
    etl_pipeline = ETLPipeline()
    import_dynamic_etls()
    spark, data = etl_pipeline.run(config=config, verbose=verbose)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", help="config file path")
    parser.add_argument("--verbose", action='store_true')
    args = parser.parse_args()
    main(args.config, args.verbose)





from pyspark.rdd import RDD
from pyspark.sql import DataFrame
from pyspark.sql.functions import split, posexplode
from pyspark.sql.functions import collect_list

from dataverse.etl.registry import register_etl

import re
import unicodedata
from typing import Union


@register_etl
def cleaning___char___normalize_whitespace(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    *args,
    **kwargs
):
    """
    Normalize the whitespaces
    - strip the leading and trailing whitespaces
    - replace all the consecutive whitespaces with a single space
        - this excludes the \n and \r

    args:
        subset (str): subset or columns to consider
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    pattern = re.compile("[^\S\r\n]+") 
    def _normalize_whitespace(row):
        row[subset] = re.sub(pattern, ' ', row[subset].strip())
        return row

    data = data.map(_normalize_whitespace)

    return data



@register_etl
def cleaning___char___remove_unprintable(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    *args,
    **kwargs
):
    """
    remove all the non-printable characters

    Code is from facebookresearch/cc_net
    https://github.com/facebookresearch/cc_net/blob/main/cc_net/text_normalizer.py

    args:
        subset (str): subset or columns to consider
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    def _remove_non_printable_char(row):
        new_lines = []
        for line in row[subset].split("\n"):
            new_lines.append(
                re.sub(
                    f"[{''.join(map(chr, list(range(0,32)) + list(range(127,160))))}]",
                    "",
                    line
                )
            )
        row[subset] = "\n".join(new_lines)
        return row

    data = data.map(_remove_non_printable_char)

    return data




"""
Data loading to Huggingface Datasets

Huggingface support spark natively!
https://huggingface.co/docs/datasets/use_with_spark
"""

import os
from datasets import Dataset
from pyspark.rdd import RDD
from pyspark.sql import DataFrame
from dataverse.etl import register_etl


@register_etl
def data_load___huggingface___ufl2hf_hub(spark, ufl, hub_path, repartition=1, *args, **kwargs):
    """
    save data to hf dataset and upload to hub
    """
    NotImplementedError()
    return None

@register_etl
def data_load___huggingface___ufl2hf(spark, ufl, save_path, repartition=1, *args, **kwargs):
    """
    save data to hf dataset and return the path
    """

    if os.path.exists(save_path):
        raise ValueError(f"save_path {save_path} already exists")

    if isinstance(ufl, RDD):
        ufl = ufl.toDF()

    assert isinstance(ufl, DataFrame), f"ufl must be RDD or DataFrame, got {type(ufl)}"

    ufl = ufl.repartition(repartition)
    hf_dataset = Dataset.from_spark(ufl)
    hf_dataset.save_to_disk(save_path)

    return save_path

@register_etl
def data_load___huggingface___ufl2hf_obj(spark, ufl, repartition=1, *args, **kwargs):
    """
    convert data to huggingface dataset object
    """
    if isinstance(ufl, RDD):
        ufl = ufl.toDF()

    assert isinstance(ufl, DataFrame), f"ufl must be RDD or DataFrame, got {type(ufl)}"

    ufl = ufl.repartition(repartition)
    hf_dataset = Dataset.from_spark(ufl)

    return hf_dataset




from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from dataverse.etl import register_etl

from typing import Union


@register_etl
def __sample___ducky___make_your_own_etl_processor(data: Union[RDD, DataFrame], *args, **kwargs):
    """
    decorator will convert this function to BaseETL class
    """
    print("make_your_own_etl_processor")
    return data




from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from typing import Union, List, Tuple
from operator import add

from dataverse.etl import register_etl


@register_etl
def utils___log___count(
    spark,
    data: Union[RDD, DataFrame],
    prev_etl_name=None,
    *args,
    **kwargs
) -> RDD[List[Tuple[str, int]]]:
    """
    Simply count the number of rows in the data

    Args:
        spark (SparkSession): SparkSession
        data (Union[RDD, DataFrame]): data to extract the nouns from
    """
    total_data = data.count()
    print('=' * 50)
    print(f"After [ {prev_etl_name} ] - Total data: {total_data}")
    print('=' * 50)

    return data




from pyspark.rdd import RDD
from pyspark.sql import DataFrame
from pyspark.sql.functions import split, posexplode
from pyspark.sql.functions import collect_list

from dataverse.etl.registry import register_etl

import re
import unicodedata
from typing import Union



def strip_accents(text: str) -> str:
    """Strips accents from a piece of text."""
    nfd = unicodedata.normalize("NFD", text)
    output = [c for c in nfd if unicodedata.category(c) != "Mn"]
    if len(output) == text:
        return text
    return "".join(output)

@register_etl
def cleaning___accent___remove(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    *args,
    **kwargs
):
    """
    strip accents from a piece of text
    - example
        - "café" -> "cafe"
        - "résumé" -> "resume"

    Code is from facebookresearch/cc_net
    https://github.com/facebookresearch/cc_net/blob/main/cc_net/text_normalizer.py

    args:
        subset (str): subset or columns to consider
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    def _strip_accents(row):
        row[subset] = strip_accents(row[subset])
        return row

    data = data.map(_strip_accents)

    return data




"""
Sampling module for data ingestion
"""

from pyspark.rdd import RDD
from dataverse.etl import register_etl

@register_etl
def utils___sampling___random(
    spark,
    ufl: RDD,
    replace=False,
    sample_n_or_frac=0.1,
    seed=42,
    *args,
    **kwargs
):
    """
    random sampling

    Args:
        ufl: ufl rdd
        replace: with replacement
        sample_n_or_frac: sample_n or fraction
        seed: seed
    """
    if isinstance(sample_n_or_frac, float):
        ufl = ufl.sample(replace, sample_n_or_frac, seed)

    # XXX: Take too long, 1M sample takes over 10 mins and didn't finish
    elif isinstance(sample_n_or_frac, int):
        ufl = ufl.takeSample(replace, sample_n_or_frac, seed)
    return ufl




"""
Load Common Crawl data from dump-id & segment files

Code is from facebookresearch/cc_net with some modifications
https://github.com/facebookresearch/cc_net

This is a migration of the code to Dataverse.
"""

import io
import os
import json
import gzip
import glob
import time
import tempfile
import requests
import functools
import warnings
import numpy as np

from pathlib import Path
from urllib.parse import urlparse
from typing import Any, List, Text, Tuple, Optional, Union, Dict, Iterable, TextIO

from dataverse.etl import register_etl
from dataverse.utils.setting import SystemSetting
from dataverse.utils.format import get_uuidv1


def parse_doc(headers: List[str], doc: List[str]) -> Optional[dict]:
    """Headers format is:
    WARC/1.0
    WARC-Type: conversion
    WARC-Target-URI: [url]
    WARC-Date: [crawldate: 2019-02-15T19:15:59Z]
    WARC-Record-ID: <urn:uuid:8865156e-d5f1-4734-9c68-4b46eaf2bb7e>
    WARC-Refers-To: <urn:uuid:340152e2-65cf-4143-b522-8ce4e2d069d7>
    WARC-Block-Digest: sha1:S3DTWCONT2L6ORTGCY2KXEZ37LNBB7V2
    Content-Type: text/plain
    Content-Length: 7743
    """
    if not headers or not doc:
        return None
    try:
        url, date, digest, length = None, None, None, None
        for header in headers:
            if header.startswith("WARC-Target-URI:"):
                url = header.split()[1]
            elif header.startswith("WARC-Date:"):
                date = header.split()[1]
            elif header.startswith("WARC-Block-Digest:"):
                digest = header.split()[1]
            elif header.startswith("Content-Length:"):
                length = int(header.split()[1])

    except Exception as e:
        # logger.warning("Can't parse header:", e, headers, doc)
        return None

    # Docs are separated by two empty lines.
    last = None
    if not doc[-1] and not doc[-2]:
        last = -2
    title, doc = doc[0], doc[1:last]

    return {
        "url": url,
        "date_download": date,
        "digest": digest,
        "length": length,
        "nlines": len(doc),
        "source_domain": urlparse(url).netloc,
        "title": title,
        "raw_content": "\n".join(doc),
    }

def group_by_docs(warc_lines: Iterable[str]) -> Iterable[dict]:
    doc: List[str] = []
    headers, read_headers = [], True
    for warc in warc_lines:
        warc = warc.strip()
        if read_headers:
            headers.append(warc)
            read_headers = warc != ""
            continue

        if warc == "WARC/1.0":
            # We reached the beginning of the new doc.
            parsed = parse_doc(headers, doc)
            if parsed is not None:
                yield parsed
            headers, doc, read_headers = [warc], [], True
            continue

        doc.append(warc)

    # Return the last document
    if doc:
        parsed = parse_doc(headers, doc)
        if parsed is not None:
            yield parsed

def _close_when_exhausted(file) -> Iterable[str]:
    with file:
        yield from file

def open_segment_file(segment: str, verbose: bool = True) -> Iterable[str]:
    """
    overwrite the open_segment function to get the WET file from the folder

    args:
        segment: path to the WET file
    """
    filename = Path(segment)
    if filename.suffix == ".gz":
        file: TextIO = gzip.open(filename, "rt")  # type: ignore
    else:
        file = open(filename, "rt")
    return _close_when_exhausted(file)

def process_segment_file(segment: str, verbose: bool = True) -> Iterable[dict]:
    for doc in group_by_docs(open_segment_file(segment, verbose=verbose)):
        doc["cc_segment"] = segment
        yield doc

def find_wet_files(directory):
    """find *.wet, *wet.gz files recursively"""
    return glob.glob(os.path.join(directory, '**/*.wet'), recursive=True) + \
           glob.glob(os.path.join(directory, '**/*.wet.gz'), recursive=True)



WET_URL_ROOT = "https://data.commoncrawl.org"
FileDescriptor = Union[Path, List[Path], str]
ReadableFileLike = Union[Iterable[str], FileDescriptor, None]

def _tmp(prefix: str = None, suffix: str = None, dir: Path = None) -> Path:
    if isinstance(prefix, Path):
        prefix = str(prefix)
    if isinstance(suffix, Path):
        suffix = str(suffix)
    _, tmp_path = tempfile.mkstemp(prefix=prefix, suffix=suffix, dir=dir)
    return Path(tmp_path)

def _yield_from(files: list) -> Iterable[str]:
    for file in files:
        yield from open_read(file)

def open_read(filename: ReadableFileLike) -> Iterable[str]:
    """Open the given file, list of files or files matching the given glob and read lines.

    `filename` is None or "-" -> reads from stdin
    `filename` is a Path / str -> interprets filename as a glob and open files matching it
    `filename` is a list -> opens sequentially all files from the list using `open_read`
    `filename` is something else -> returns the object wrapped in a `nullcontext`
        This allows to pass already openened files or iterables.

    `open_read` will decompress gzip files, given they have ".gz" suffix.
    """
    if filename is None:
        return sys.stdin

    if isinstance(filename, list):
        assert isinstance(filename[0], Path)
        if len(filename) == 0:
            return []
        if len(filename) > 1:
            return _yield_from(filename)
        filename = tp.cast(Path, filename[0])
    if isinstance(filename, str):
        if filename.startswith("http://") or filename.startswith("https://"):
            return open_remote_file(filename)

        filename = Path(filename)
    if not isinstance(filename, Path):
        # we might have received an iterable, return it unmodified.
        return filename  # type: ignore

    # Expand glob patterns only when reading
    files = [Path(f) for f in sorted(glob.glob(str(filename)))]
    if len(files) > 1:
        return _yield_from(files)
    if len(files) == 1:
        filename = files[0]

    assert isinstance(filename, Path)

    if filename.suffix == ".gz":
        file: TextIO = gzip.open(filename, "rt")  # type: ignore
    else:
        file = open(filename, "rt")

    return _close_when_exhausted(file)


def request_get_content(url: str, n_retry: int = 3, verbose: bool = True) -> bytes:
    """Retrieve the binary content at url.

    Retry on connection errors.
    """
    t0 = time.time()

    if verbose:
        # TODO: Logging will be activated later
        # logging.info(f"Starting download of {url}")
        print(f"Starting download of {url}")

    for i in range(1, n_retry + 1):
        try:
            with requests.Session() as session:
                r = session.get(url)
                r.raise_for_status()
            break
        except requests.exceptions.RequestException as e:
            # Sleep and try again on error, unless it's a 404.
            message = e.args[0] if isinstance(e.args[0], str) else ""
            if i == n_retry or "Client Error" in message:
                raise e
            warnings.warn(
                f"Swallowed error {e} while downloading {url} ({i} out of {n_retry})"
            )
            time.sleep(10 * 2 ** i)


    if verbose:
        dl_time = time.time() - t0
        dl_speed = len(r.content) / dl_time / 1024
        # logging.info(
        #     f"Downloaded {url} [{r.status_code}] took {dl_time:.0f}s ({dl_speed:.1f}kB/s)"
        # )
        print(f"Downloaded {url} [{r.status_code}] took {dl_time:.0f}s ({dl_speed:.1f}kB/s)")

    return r.content


def open_remote_file(url: str, cache: Path, verbose: bool = True) -> Iterable[str]:
    """
    Download the files at the given url to memory and opens it as a file.
    Assumes that the file is small, and fetch it when this function is called.
    """
    if cache and cache.exists():
        return open_read(cache)

    # TODO: open the remote file in streaming mode.
    # The hard part is that we need to write the content on disk at the same time,
    # to implement disk caching.
    raw_bytes = request_get_content(url, verbose=verbose)
    content = io.BytesIO(raw_bytes)
    if url.endswith(".gz"):
        f: TextIO = gzip.open(content, mode="rt")  # type: ignore
    else:
        f = io.TextIOWrapper(content)

    try:
        # The file might have been created even not fully downloaded/written
        # so make sure tmp_cache is deleted when the program exits.
        # and only replace the cache file when the download is complete.
        if cache and not cache.exists():
            tmp_cache = _tmp(cache)
            tmp_cache.write_bytes(raw_bytes)
            if not cache.exists():
                tmp_cache.replace(cache)
    finally:
        if tmp_cache.exists():
            tmp_cache.unlink()

    return _close_when_exhausted(f)


def cc_wet_paths_url(dump_id: str) -> str:
    return "/".join([WET_URL_ROOT, "crawl-data", "CC-MAIN-" + dump_id, "wet.paths.gz"])

def segment_url(segment: str):
    return "/".join((WET_URL_ROOT, segment))

def cc_segment_urls(dump_id: str, cache_dir: Path, verbose: bool = True) -> List[str]:
    wet_paths = cc_wet_paths_url(dump_id)
    wet_paths_cache = cache_dir / f"wet_{dump_id}.paths.gz"
    f = open_remote_file(wet_paths, cache=wet_paths_cache, verbose=verbose)
    return [segment.strip() for segment in f]

def open_segment_url(segment: str, cache_dir: Path, verbose: bool = True) -> Iterable[str]:
    url = segment_url(segment)
    file: Optional[Path] = None
    if cache_dir:
        file = cache_dir / segment.split("/")[-1]

    return open_remote_file(url, cache=file, verbose=verbose)

def process_segment_url(segment: str, cache_dir: Path, verbose: bool = True) -> Iterable[str]:
    for doc in group_by_docs(open_segment_url(segment, cache_dir, verbose=verbose)):
        doc["cc_segment"] = segment
        yield doc




@register_etl
def data_ingestion___common_crawl___wet2raw(
    spark,
    wet_path: str,
    segment_n: int = -1,
    repartition=20,
    seed: int = 42,
    verbose=True,
    *args,
    **kwargs
):
    """
    load WET to raw format as dict

    [ what is WET? ]
    - WET files which store extracted plain text from the data stored in the WARC‍

    Args:
        spark (SparkSession): spark session
        wet_path (str): path to the WET folder includes WET format files
            - this search recursively, so you don't need to specify the path to each WET file
            - this search for all the *.wet, *.gz files in the folder
        segment_n (int): the number of segments to load, this is sampling
            - one segment is about 1GB
            - (default) set as -1 if you wanna load all the segments
        repartition (int): the number of partitions
        seed (int): random seed
        verbose (bool): whether to print the information of the dataset
    """
    wet_paths = find_wet_files(wet_path)
    if segment_n > 0 and segment_n < len(wet_paths):
        np.random.seed(seed)
        wet_paths = np.random.choice(wet_paths, size=segment_n, replace=False)

    rdd = spark.sparkContext.parallelize(wet_paths)
    rdd = rdd.flatMap(functools.partial(process_segment_file, verbose=verbose))
    rdd = rdd.repartition(repartition)

    return rdd


@register_etl
def data_ingestion___common_crawl___dump2raw(
    spark,
    dump: str,
    segment_n: int = -1,
    repartition=20,
    use_cache: bool = True,
    cache_dir: str = None,
    seed: int = 42,
    verbose=True,
    *args,
    **kwargs
):
    """
    Args:
        spark (SparkSession): spark session
        dump (str): dump id of the Common Crawl
            - ex) 2023-23
        segment_n (int): the number of segments to load, this is sampling
            - one segment is about 1GB
            - (default) set as -1 if you wanna load all the segments
        repartition (int): the number of partitions
        use_cache (bool): whether to use the cache
            - set as False if you wanna save disk space, because the cache is huge
            - one WET dump is about 10TB
        cache_dir (str): cache path to save the dataset
        seed (int): random seed
        verbose (bool): whether to print the information of the dataset
    """
    if use_cache:
        if cache_dir is None:
            # save the parquet at package root path
            cache_dir = SystemSetting().CACHE_DIR
            cache_dir = f"{cache_dir}/.cache/dataverse/dataset/common_crawl_{dump}"
        else:
            cache_dir = f"{cache_dir}/common_crawl_{dump}"
    else:
        cache_dir = None

    if not isinstance(cache_dir, Path):
        cache_dir = Path(cache_dir)

    # if cache dir exist creat one
    if cache_dir and not cache_dir.exists():
        cache_dir.mkdir(parents=True)

    wet_urls = cc_segment_urls(dump, cache_dir, verbose=verbose)

    if segment_n > 0 and segment_n < len(wet_urls):
        np.random.seed(seed)
        wet_urls = np.random.choice(wet_urls, size=segment_n, replace=False)

    rdd = spark.sparkContext.parallelize(wet_urls)
    rdd = rdd.flatMap(functools.partial(
        process_segment_url,
        cache_dir=cache_dir,
        verbose=verbose,
    ))
    rdd = rdd.repartition(repartition)

    return rdd


def convert_bytes(data):
    if isinstance(data, bytes):
        return data.decode()
    if isinstance(data, dict):
        return {convert_bytes(key): convert_bytes(value) for key, value in data.items()}
    if isinstance(data, list):
        return [convert_bytes(element) for element in data]
    return data

@register_etl
def data_ingestion___common_crawl___raw2ufl(spark, data, *args, **kwargs):
    """
    convert raw format to ufl with custom template
    """
    def templatev1(data):
        new_data = {}
        new_data['id'] = get_uuidv1()
        new_data['name'] = 'common_crawl'
        new_data['text'] = (
            f"{data.get('raw_content', None)}"
        )
        new_data['meta'] = json.dumps(
            convert_bytes({
                'title': data.get('title', None),
                'url': data.get('url', None),
                'date_download': data.get('date_download', None),
                'digest': data.get('digest', None),
                'length': data.get('length', None),
                'nlines': data.get('nlines', None),
                'source_domain': data.get('source_domain', None),
                'cc_segment': data.get('cc_segment', None),
            })
        )
        return new_data

    data = data.map(lambda x: templatev1(x))

    return data



"""
base class to support the registration of the ETL classes
"""

import os
import abc
import inspect
import importlib.util

from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from typing import Union
from dataverse.utils.setting import SystemSetting


# TODO: If you add category directories, add them here too
# _sample is a special directory that is not imported

# This is where you choose what categories to register
ETL_CATEGORIES = [
    'data_ingestion',
    'decontamination',
    'deduplication',
    'bias',
    'toxicity',
    'cleaning',
    'pii',
    'quality',
    'data_load',
    'utils',
]

IGNORE_FILES = [
    '__init__.py',
]

def auto_register(etl_categories=ETL_CATEGORIES):
    """
    This will automatically register all ETLs to the registry
    """
    etl_path = os.path.dirname(os.path.abspath(__file__))
    for etl_category in etl_categories:
        
        # Get the files(sub-categories) in the category
        category_path = os.path.join(etl_path, etl_category)
        files = os.listdir(category_path)

        # Filter out non-Python files
        files = [f for f in files if f.endswith('.py')]

        # Dynamically import all Python files in the directory
        for file in files:
            if file in IGNORE_FILES:
                continue

            file_path = os.path.join(category_path, file)

            # Remove .py at the end
            module_name = file[:-3]

            spec = importlib.util.spec_from_file_location(module_name, file_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)


# To avoid circular dependency
class ETLStructure: ...


class ETLRegistry:
    """Singleton class to register the ETL classes"""
    _initialized = False

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(ETLRegistry, cls).__new__(cls)
        return cls.instance

    def __init__(self):
        """
        when the class is initialized, this is called everytime
        regardless of the singleton. So adding the flag to check
        """
        if self._initialized:
            return
        self._initialized = True
        self._registry = {}
        self._status = {}
        auto_register()


    def __len__(self):
        return len(self._registry.keys())

    def __repr__(self):
        return self._convert_to_report_format(self._status)

    def __str__(self):
        return self.__repr__()

    def reset(self):
        """
        reset the registry
        """
        self._registry = {}

    def register(self, key: str, etl: ETLStructure):
        """
        register the etl
        """
        # the key should be the format of the following
        # ==================================================
        # <etl_type>___<file_key>___<etl_key>
        # <category>___<sub_category>___<etl_key>
        # ==================================================
        # 1. is it all lowercase
        # 2. is it separated by ___
        # 3. it should have 2 layers of category
        if not key.islower():
            raise ValueError(f"The key [ {key} ] should be all lowercase")
        if "___" not in key:
            raise ValueError(f"The key [ {key} ] should be separated by ___")
        if len(key.split("___")) != 3:
            raise ValueError(f"The key [ {key} ] should have 2 layers of category")

        # all the etl should be the subclass of ETLStructure
        if not issubclass(etl, ETLStructure):
            raise TypeError(f"ETL class should be subclass of ETLStructure not {etl}")

        # register already exists
        if key in self._registry:
            raise KeyError(f"The key [ {key} ] is already registered")

        self._registry[key] = etl
        self._update_status(key=key)

    def _update_status(self, key: str):
        category, sub_category, _ = key.split('___')
        if category not in self._status:
            self._status[category] = {}

        if sub_category not in self._status[category]:
            self._status[category][sub_category] = [key]
        else:
            self._status[category][sub_category].append(key)

    def search(self, category=None, sub_category=None):
        """
        search the etl

        printing all the information is fixed as default
        - print_sub_category: print the sub-category
        - print_etl_name: print the etl name
        """
        status = self._status

        filtered_status = {}
        if category is not None:
            assert type(category) != list, 'we do not support list search for category'
            assert type(category) == str, 'category must be a string'
            if sub_category is None:
                filtered_status[category] = status[category]
            else:
                assert type(sub_category) != list, 'we do not support list search for sub-category'
                assert type(sub_category) == str, 'sub_category must be a string'
                filtered_status[category] = {sub_category: status[category][sub_category]}
        else:
            if sub_category is not None:
                raise ValueError('sub-category cannot be specified without category')
            filtered_status = status

        return self._convert_to_report_format(
            filtered_status,
            print_sub_category=True,
            print_etl_name=True,
        )

    def _convert_to_report_format(
        self,
        status,
        print_sub_category=False,
        print_etl_name=False,
    ):
        """
        convert status to report format

        This includes the number of ETLs in each category and sub-category
        and depending on the options, it can include the name of the ETLs

        Args:
            status (dict): the status from `search`
        """
        # count the number of etls
        stats = {}
        total = 0
        categories = list(status.keys())
        for category in categories:
            if category not in stats:
                stats[category] = {}
                stats[category]['__total__'] = 0

            sub_categories = list(status[category].keys())
            for sub_category in sub_categories:
                sub_n = len(status[category][sub_category])
                stats[category][sub_category] = sub_n
                stats[category]['__total__'] += sub_n
                total += sub_n

        # convert to the report format
        infos = []

        infos.append('=' * 50)
        infos.append(f"Total [ {total} ]")
        infos.append('=' * 50)

        for category in categories:
            infos.append(f"{category} [ {stats[category]['__total__']} ]")
            sub_categories = list(status[category].keys())

            if print_sub_category:
                for sub_category in sub_categories:   
                    infos.append(f"{' ' * 4}- {sub_category} [ {stats[category][sub_category]} ]")

                    if print_etl_name:
                        for etl in status[category][sub_category]:
                            infos.append(f"{' ' * 8}- {etl}")

        return '\n'.join(infos)

    def get(self, key: str) -> ETLStructure:
        """
        get the etl
        """
        # the key should be the format of the following
        # ==================================================
        # <etl_type>___<file_key>___<etl_key>
        # <category>___<sub_category>___<etl_key>
        # ==================================================
        # 1. is it all lowercase
        # 2. is it separated by ___
        # 3. it should have 2 layers of category
        if not key.islower():
            raise ValueError(f"The key [ {key} ] should be all lowercase")
        if "___" not in key:
            raise ValueError(f"The key [ {key} ] should be separated by ___")
        if len(key.split("___")) != 3:
            raise ValueError(f"The key [ {key} ] should have 2 layers of category")

        if key not in self._registry:
            raise KeyError(f"The key {key} is not registered")

        return self._registry[key]

    def get_all(self):
        """
        get all the etls
        """
        return list(self._registry.values())

class ETLAutoRegistry(abc.ABCMeta, type):
    def __new__(cls, name, bases, attrs):
        """
        Metaclass to register the ETL classes automatically to the registry
        """
        # singleton registry
        new_class = super().__new__(cls, name, bases, attrs)

        # BaseETL is base class and should not be registered
        # Another reason is BaseETL is not initialized yet before `__new__` is done but 
        # the registry will verify the class is subclass of BaseETL and raise error
        # because BaseETL is not initialized yet :)
        if name != 'BaseETL':
            if "__file_path__" not in attrs:
                raise TypeError("Direct inheritance from BaseETL not allowed. Use @register_etl decorator.")

            registry = ETLRegistry()
            registry.register(key=name, etl=new_class)

        return new_class


class BaseETL(ETLStructure, metaclass=ETLAutoRegistry):
    """
    spark ETL
    """
    @abc.abstractmethod
    def run(self, data: Union[RDD, DataFrame], *args, **kwargs):
        """
        run the preprocessing
        """
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        """
        call the method to do the preprocessing
        """
        return self.run(*args, **kwargs)


def add_self(func):
    """
    Decorator to add self to the function
    intent is to make the function as a method
    """
    def wrapper(self, *args, **kwargs):
        return func(*args, **kwargs)
    return wrapper

def register_etl(func):
    """
    Decorator to register a function as an ETL

    one downside is that the function cannot leverage the `self`
    if you need to use `self` directly inherit the BaseETL

    About Attributes:
    - __file_path__ (str): the file path of the function where it is defined
    - __etl_dir__ (bool): if the file is in the etl directory
        - if not, it means it's dynamically registered user-defined ETL
    """
    ETL_DIR = os.path.join(SystemSetting().DATAVERSE_HOME, 'etl')
    etl_file_path = inspect.getfile(func)

    # I know using class name without snake case is awkward
    # but I want to keep the class name as it is and user won't know it
    etl_cls = type(
        func.__name__,
        (BaseETL,),
        {
            "run": add_self(func),
            "__file_path__": etl_file_path,
            "__etl_dir__": etl_file_path.startswith(ETL_DIR),
        }
    )

    return etl_cls



"""
ETL Interface 
----------------------
user will be interacting with this interface
"""

import time
import boto3

from typing import Union
from pathlib import Path
from omegaconf import OmegaConf
from omegaconf import DictConfig

from pyspark.rdd import RDD
from pyspark.sql import DataFrame
from pyspark.conf import SparkConf
from pyspark.sql import SparkSession

from dataverse.config import Config
from dataverse.etl import ETLRegistry
from dataverse.utils.setting import SystemSetting
from dataverse.utils.api import aws_check_credentials
from dataverse.utils.api import EMRManager
from dataverse.utils.api import AWSClient


class ETLPipeline:
    """
    ETL Pipeline
    """
    def __init__(self):
        self.registry = ETLRegistry()

    def __len__(self):
        return len(self.registry)

    def status(self):
        """
        get the status of the registry. don't show you detailed information

        [ info type ]
        - category

        usage:
            >>> etl_pipeline.status()
        """
        print("If you need details of ETL Registry use `etl_pipeline.search()`")
        return str(self.registry)

    def search(self, category=None, sub_category=None):
        """
        get detailed status of the registry by searching

        [ info type ]
        - category
        - sub_category
        - etl_name

        usage:
            # return every etl
            >>> etl_pipeline.search()

            # only selected category
            >>> etl_pipeline.search('data_ingestion')
            >>> etl_pipeline.search(category='data_ingestion')

            # only selected category & sub_category
            >>> etl_pipeline.search('data_ingestion', 'ufl')
            >>> etl_pipeline.search(category='data_ingestion', sub_category='ufl')
        """
        return self.registry.search(category=category, sub_category=sub_category)

    def get(self, key):
        """get ETL class from registry"""
        return self.registry.get(key=key)

    def setup_spark_conf(self, config, verbose=False):
        """
        AWS credential setting log is not influenced by the verbose by design
        """

        # TODO: add more spark configurations
        spark_conf = SparkConf()
        spark_conf.set('spark.master', config.spark.master)
        spark_conf.set('spark.app.name', config.spark.appname)
        spark_conf.set('spark.driver.memory', config.spark.driver.memory)
        spark_conf.set('spark.driver.maxResultSize', config.spark.driver.maxResultSize)
        spark_conf.set('spark.executor.memory', config.spark.executor.memory)
        spark_conf.set('spark.local.dir', config.spark.local.dir)
        spark_conf.set('spark.ui.port', config.spark.ui.port)

        # AWS S3 Support
        if aws_check_credentials(verbose=verbose):
            session = boto3.Session()
            credentials = session.get_credentials()

            spark_conf.set('spark.hadoop.fs.s3a.access.key', credentials.access_key)
            spark_conf.set('spark.hadoop.fs.s3a.secret.key', credentials.secret_key)
            spark_conf.set('spark.hadoop.fs.s3a.impl', 'org.apache.hadoop.fs.s3a.S3AFileSystem')

            hadoop_ver = SystemSetting().get('HADOOP_VERSION')
            spark_conf.set(
                'spark.jars.packages',
                (
                    f'org.apache.hadoop:hadoop-aws:{hadoop_ver}'
                    f',com.amazonaws:aws-java-sdk-bundle:1.12.592'
                )
            )

            # check if the credentials are temporary or not
            try:
                spark_conf.set('spark.hadoop.fs.s3a.session.token', credentials.token)
                spark_conf.set(
                    'spark.hadoop.fs.s3a.aws.credentials.provider',
                    "org.apache.hadoop.fs.s3a.TemporaryAWSCredentialsProvider"
                ) # this is for temporary credentials
                print('spark conf is set with [ temporary ] S3 credentials')
            except:
                print('spark conf is set with [ permanent ] S3 credentials')

        else:
            print('[ No AWS Credentials Found] - Failed to set spark conf for S3')

        return spark_conf


    def sample(
        self,
        n=100,
        config=None,
        sample_etl="data_ingestion___test___generate_fake_ufl",
        verbose=False,
    ):
        """
        get spark session and sample data

        use this function to test the ETL pipeline quickly without config

        args:
            n (int): the number of data to generate
            config (Union[str, dict, OmegaConf]): config for the etl
                - config is not necessary
            sample_etl (str): the name of the sample ETL process
            verbose (bool): if True, print the status
        """
        if config is None:
            config = Config.default()
        else:
            config = Config.load(config)
            config = Config.set_default(config)

            # remove all the ETL processes
            config.etl = []

        config.etl.append({'name': sample_etl, 'args': {'n': n}})
        if verbose:
            print('=' * 50)
            print('[ Configuration ]')
            print(OmegaConf.to_yaml(config))
            print('=' * 50)

        spark_conf = self.setup_spark_conf(config, verbose=verbose)
        spark = SparkSession.builder.config(conf=spark_conf).getOrCreate()
        if verbose:
            print('=' * 50)
            print('[ Spark Final Configuration ]')
            print(OmegaConf.to_yaml(spark_conf.getAll()))
            print('=' * 50)

        sample_etl_class = self.get(key=sample_etl)
        data = sample_etl_class()(spark, n=n, etl_name=sample_etl)

        if verbose:
            print((
                f"{'=' * 50}\n"
                "[ SAMPLE MODE ]\n"
                f"{'=' * 50}\n"
                "This is a quick way to get the sample data for testing or debugging w/o config.\n"
                "If you want to test the ETL pipeline with your own data, please use `run` w/ config.\n"
                f"{'=' * 50}\n"
                "=> spark, data = etl_pipeline.sample()\n"
                "=> data = data.map(add awesome duck to column)\n"
                f"{'=' * 50}\n"
            ))

        return spark, data


    def run(
        self,
        config: Union[str, dict, DictConfig, OmegaConf, Path],
        verbose=False,
        cache=False,
        emr=False,
        *args,
        **kwargs,
    ):
        """
        Args:
            config (Union[str, dict, OmegaConf]): config for the etl
                - str: path to the config file
                - dict: config dict
                - OmegaConf: config object
            verbose (bool): if True, print the status of the etl pipeline
                - the verbose will be applied to the ETL process as well
                - ETL process `verbose` takes precedence over this
            cache (bool): cache every stage of the ETL process
            emr (bool): if True, run the ETL process on EMR
        """
        # ================ [ EMR ] ===================
        if emr:
            return self.run_emr(
                config,
                verbose=verbose,
                cache=cache,
                *args,
                **kwargs,
            )

        # =============== [ Set Config ] ==================
        # mainly this is to fill the missing config args with default
        config = Config.load(config)
        config = Config.set_default(config)
        if verbose:
            print('=' * 50)
            print('[ Configuration ]')
            print(OmegaConf.to_yaml(config))
            print('=' * 50)

        # ================ [ Set Spark ] ===================
        spark_conf = self.setup_spark_conf(config, verbose=verbose)
        spark = SparkSession.builder.config(conf=spark_conf).getOrCreate()
        if verbose:
            print('=' * 50)
            print('[ Spark Final Configuration ]')
            print(OmegaConf.to_yaml(spark_conf.getAll()))
            print('=' * 50)

        # ================= [ Run ETL ] ====================
        # [ Load RDD/DataFrame ] - data ingestion
        # [ Preprocessing ]
        # [ Save RDD/DataFrame ] - data load
        etl_configs = config.etl
        total_etl_n = len(etl_configs)

        # [switch] is the ETL process ended or not
        # if not, spark session & data will be returned to continue
        IS_ETL_FINISHED = True

        data = None
        prev_etl_name = None
        prev_data = None # for caching
        for etl_i, etl_config in enumerate(etl_configs):
            # etl_config.name format
            # =====>[ etl_cate___etl_sub_cate___etl_name ]
            etl_name = etl_config.name
            etl_category = etl_name.split('___')[0]
            etl_class = self.get(key=etl_name)
            
            # instantiate etl class
            etl_instance = etl_class()

            # this is middle creator mode
            # if the last ETL process is not data load  
            if etl_i == total_etl_n - 1 and etl_category != 'data_load':
                if verbose:
                    print((
                        f"{'=' * 50}\n"
                        "[ DEBUG MODE ]\n"
                        f"{'=' * 50}\n"
                        f"Last ETL process was assigned for [ {etl_category} ]\n"
                        "Spark session will not be stopped and will be returned\n"
                        "If this is not intended, please assign [ data_load ] at the end.\n"
                        f"{'=' * 50}\n"
                        "Example:\n"
                        "=> spark, data = etl_pipeline.run(config)\n"
                        "=> data = data.map(add awesome duck to column)\n"
                        f"{'=' * 50}\n"
                    ))
                IS_ETL_FINISHED = False

            # when args is not defined, set it to empty dict
            if 'args' in etl_config:
                args = etl_config.args
            else:
                args = {}

            # if verbose is not defined, set it same to the pipeline
            if 'verbose' not in args:
                args['verbose'] = verbose

            # `etl_name` is passed to args for tracking
            if etl_i == 0:
                data = etl_instance(spark, **args, etl_name=etl_name, prev_etl_name=None)
            else:
                data = etl_instance(spark, data, **args, etl_name=etl_name, prev_etl_name=prev_etl_name)

            # cache the data
            if cache:
                if prev_data is not None:
                    prev_data.unpersist()
                data.cache()
                prev_data = data

            prev_etl_name = etl_name


        # =============== [ Stop Spark ] ==================
        if IS_ETL_FINISHED:
            spark.stop()
            if verbose:
                print('=' * 50)
                print('[ Spark Successfully Done ]')
                print('=' * 50)

        return spark, data

    def run_emr(
        self,
        config: Union[str, dict, DictConfig, OmegaConf, Path],
        verbose=False,
        cache=False,
        *args,
        **kwargs,
    ):
        """
        Args:
            config (Union[str, dict, OmegaConf]): config for the etl
                - str: path to the config file
                - dict: config dict
                - OmegaConf: config object
            verbose (bool): if True, print the status of the etl pipeline
                - the verbose will be applied to the ETL process as well
                - ETL process `verbose` takes precedence over this
            cache (bool): cache every stage of the ETL process

        Returns:
            None, Config:
                - None for spark session
                - Config for the config
                    - originally data is returned, but it is not necessary for EMR
                    - so rather data, config is returned
        """
        if not aws_check_credentials(verbose=verbose):
            raise ValueError('AWS EMR requires AWS credentials')

        # =============== [ Set Config ] ==================
        config = Config.load(config)
        config = Config.set_default(config, emr=True)

        # EMR resource manager - yarn
        config.spark.master = 'yarn'

        # reset local_dir for EMR cluster
        config.spark.local.dir = '/tmp'

        # ================ [ EMR ] ===================
        # NOTE: config will be auto-updated by EMR Manager
        emr_manager = EMRManager()

        try:
            # EMR cluster launch
            emr_manager.launch(config)

            if verbose:
                print('=' * 50)
                print('[ Configuration ]')
                print(OmegaConf.to_yaml(config))
                print('=' * 50)

            # EMR cluster environment setup & run spark
            step_id = emr_manager.run(config, verbose=verbose)

            # wait until EMR cluster step is done
            emr_manager.wait(config, step_id)

            # EMR Cluster terminate
            # XXX: after EMR cluster is terminated, and confirmed by waiter
            #      there is still a chance that the cluster is not terminated and cause error
            #       - DependencyViolation (which depends on terminated cluster)
            # FIXME: this is a temporary solution, need to find a better way to handle this
            RETRY_TERMINATE = 5
            for _ in range(RETRY_TERMINATE):
                try:
                    emr_manager.terminate(config)
                    break
                except AWSClient().ec2.exceptions.ClientError as e:
                    if e.response['Error']['Code'] == 'DependencyViolation':
                        print('DependencyViolation - retrying to terminate EMR cluster')
                        time.sleep(5)
                    else:
                        raise e
                except Exception as e:
                    raise e

        # ctrl + c
        except KeyboardInterrupt:
            print('KeyboardInterrupt - terminating EMR cluster')
            emr_manager.terminate(config)
            raise KeyboardInterrupt
        except Exception as e:
            print('Exception - terminating EMR cluster')
            emr_manager.terminate(config)
            raise e

        return None, config



"""
Code is from ChenghaoMou/text-dedup
https://github.com/ChenghaoMou/text-dedup/blob/main/text_dedup/minhash_spark.py

This is a migration of the code to Dataverse.
"""

import pyspark
from pyspark.rdd import RDD
from pyspark.sql import DataFrame
from pyspark.sql import functions as F
from dataverse.etl.registry import register_etl
from typing import Union, List, Tuple, Any, Text

import re
import sys
import hashlib
import struct
import numpy as np

from operator import add
from itertools import tee
from scipy.integrate import quad as integrate


NON_ALPHA = re.compile(r"\W", re.UNICODE)
DTYPE = np.uint32
MAX_HASH = 4_294_967_295
MOD_PRIME = 4_294_967_291


# region: Connected Components in MapReduce and Beyond, 2014
def generate_edges(nodes: List[int]) -> List[Tuple[int, int]]:
    """
    Generate edges from a cluster. Instead of generating N^2 edges, we only need all nodes align to a single node, since
    we will be running connected components on the edges later.

    Parameters
    ----------
    nodes : List[int]
        The list of nodes in the cluster.

    Returns
    -------
    List[Tuple[int, int]]
        The list of edges.

    Examples
    --------
    >>> generate_edges([1, 2, 3])
    [(2, 1), (3, 1)]
    """
    if len(nodes) <= 1:
        return []

    min_node = min(nodes)
    return [(n, min_node) for n in nodes if n != min_node]


def small_star(edges):
    def small_star_map(edge):
        x, y = edge
        if y <= x:
            return (x, y)
        else:
            return (y, x)

    def small_star_reduce(x):
        node, neighbors = x
        nodes = neighbors + [node]
        min_node = min(nodes)
        new_edges = set((neighbor, min_node) for neighbor in nodes if (neighbor <= node and neighbor != min_node))
        change = len(new_edges.difference(set([(node, neighbor) for neighbor in neighbors])))
        return (list(new_edges), change)

    neighbors = edges.map(small_star_map).groupByKey().map(lambda x: (x[0], list(set(x[1]))))
    edges_with_change = neighbors.map(small_star_reduce).cache()

    # no duplicate edges
    if edges_with_change.isEmpty():
        total_change = 0
    else:
        total_change = edges_with_change.map(lambda x: x[1]).reduce(add)

    edges = edges_with_change.flatMap(lambda x: x[0])
    edges_with_change.unpersist()

    return edges, total_change


def large_star(edges):
    def large_star_map(edge):
        if edge[0] == edge[1]:
            return [(edge[0], edge[1])]
        return [(edge[0], edge[1]), (edge[1], edge[0])]

    def large_star_reduce(x):
        node, neighbors = x
        nodes = neighbors + [node]
        min_node = min(nodes)
        new_edges = set((neighbor, min_node) for neighbor in (neighbors + [node]) if (neighbor > node))
        change = len(new_edges.difference(set([(node, neighbor) for neighbor in neighbors])))
        return list(new_edges), change

    neighbors = edges.flatMap(large_star_map).groupByKey().map(lambda x: (x[0], list(set(x[1]))))
    edges_with_change = neighbors.map(large_star_reduce).cache()

    # no duplicate edges
    if edges_with_change.isEmpty():
        total_change = 0
    else:
        total_change = edges_with_change.map(lambda x: x[1]).reduce(add)

    edges = edges_with_change.flatMap(lambda x: x[0])
    edges_with_change.unpersist()
    return edges, total_change


def alternating_algo(edges, max_iteration: int) -> Tuple[Any, bool, int]:
    prev_lchanges: int = sys.maxsize
    prev_schanges: int = sys.maxsize
    curr_iteration: int = 0

    while max_iteration:
        edges, curr_lchanges = large_star(edges)
        edges, curr_schanges = small_star(edges)

        if (curr_lchanges == prev_lchanges and curr_schanges == prev_schanges) or (
            curr_schanges == 0 and curr_lchanges == 0
        ):
            return edges, True, curr_iteration

        prev_lchanges = curr_lchanges
        prev_schanges = curr_schanges
        curr_iteration += 1
        max_iteration -= 1

    return edges, False, curr_iteration

# endregion


# region: Hashing
def ngrams(sequence: List[Text], n: int, min_length: int = 5):
    """
    Return the ngrams generated from a sequence of items, as an iterator.

    This is a modified version of nltk.util.ngrams.

    Parameters
    ----------
    sequence : List[Text]
        The sequence of items.
    n : int
        The length of each ngram.
    min_length : int, optional
        The minimum length of each ngram, by default 5

    Returns
    -------
    iterator
        The ngrams.

    Examples
    --------
    >>> list(ngrams(["a", "b", "c", "d"], 2, min_length=1))
    [('a', 'b'), ('b', 'c'), ('c', 'd')]
    >>> list(ngrams(["a", "b", "c", "d"], 2, min_length=5))
    []
    >>> list(ngrams(["a", "b"], 3, min_length=1))
    [('a', 'b')]
    """
    if len(sequence) < min_length:
        return []
    if len(sequence) < n:
        return [tuple(sequence)]
    iterables = tee(iter(sequence), n)
    for i, sub_iterable in enumerate(iterables):
        for _ in range(i):
            next(sub_iterable, None)
    return zip(*iterables)


def sha1_hash32(data):
    """
    Directly taken from datasketch package to avoid abstraction.

    Parameters
    ----------
    data : bytes

    Returns
    -------
    int
        The first 4 bytes (32 bits) of the SHA1 hash of the input data.

    Examples
    --------
    >>> sha1_hash32(b"hello")
    499578026
    >>> bin(sha1_hash32(b"hello"))
    '0b11101110001101111010010101010'
    >>> sha1_hash32(b"hello world").bit_length()
    30
    """
    return struct.unpack("<I", hashlib.sha1(data).digest()[:4])[0]


def generate_hash_values(
    content: str,
    idx: int,
    num_perm: int,
    ngram_size: int,
    min_length: int,
    hashranges: List[Tuple[int, int]],
    permutations: Tuple[np.ndarray, np.ndarray],
) -> List[Tuple[int, bytes, int]]:
    """
    Generate the MinHashLSH values for a given document.

    Parameters
    ----------
    content : str
        The content of the document.
    idx : int
        The index of the document.
    num_perm : int
        The number of permutations.
    ngram_size : int
        The size of the n-grams.
    min_length : int
        The minimum number of tokens in a document.
    hashranges : list
        The ranges of offsets for each hash value.
    permutations : Tuple[np.ndarray, np.ndarray]
        The permutations for the hash values.

    Returns
    -------
    List[Tuple[int, bytes, int]]
        The list of (band_idx, hash value, idx) for the document.

    Examples
    --------
    >>> content = "hello world"
    >>> idx = 0
    >>> num_perm = 250
    >>> ngram_size = 1
    >>> hashranges = [(i, i + 25) for i in range(0, 250, 25)]
    >>> PERMUTATIONS = (
    ...     RNG.randint(1, MOD_PRIME, size=(num_perm,), dtype=DTYPE),
    ...     RNG.randint(0, MOD_PRIME, size=(num_perm,), dtype=DTYPE),
    ... )
    >>> res = generate_hash_values(content, idx, num_perm, ngram_size, 0, hashranges, PERMUTATIONS)
    >>> len(res)
    10
    >>> sum(len(h) for _, h, _ in res) == len(res) * 25 * np.dtype(DTYPE).itemsize
    True
    """
    tokens = {" ".join(t).encode("utf-8") for t in ngrams(NON_ALPHA.split(content.lower()), ngram_size, min_length)}
    a, b = permutations
    hv = np.array([sha1_hash32(token) for token in tokens], dtype=DTYPE)
    phv = np.bitwise_and(((hv * np.tile(a, (len(tokens), 1)).T).T + b) % MOD_PRIME, MAX_HASH)  # type: ignore
    hash_values = np.vstack([phv, np.ones(num_perm, dtype=DTYPE) * MAX_HASH]).min(axis=0)
    Hs = [bytes(hash_values[start:end].byteswap().data) for start, end in hashranges]
    return [(band_idx, H, idx) for band_idx, H in enumerate(Hs)]


# endregion

# region: MinHashLSH
def optimal_param(
    threshold: float,
    num_perm: int,
    false_positive_weight: float = 0.5,
    false_negative_weight: float = 0.5,
):
    """
    Compute the optimal `MinHashLSH` parameter that minimizes the weighted sum
    of probabilities of false positive and false negative, taken from datasketch.

    Parameters
    ----------
    threshold : float
        The threshold for similarity.
    num_perm : int
        The number of permutations.
    false_positive_weight : float
        The weight of false positive.
    false_negative_weight : float
        The weight of false negative.

    Returns
    -------
    Tuple[int, int]
        The optimal `b` and `r` parameters.
        The number of bands, and the number of rows per band respectively.

    Examples
    --------
    >>> optimal_param(0.7, 256)
    (25, 10)
    """

    def false_positive_area(threshold: float, b: int, r: int):
        """Source: `datasketch.lsh`"""

        def area(s):
            return 1 - (1 - s ** float(r)) ** float(b)

        a, _ = integrate(area, 0.0, threshold)
        return a

    def false_negative_area(threshold: float, b: int, r: int):
        """Source: `datasketch.lsh`"""

        def area(s):
            return 1 - (1 - (1 - s ** float(r)) ** float(b))

        a, _ = integrate(area, threshold, 1.0)
        return a

    min_error = float("inf")
    opt = (0, 0)
    for b in range(1, num_perm + 1):
        max_r = int(num_perm / b)
        for r in range(1, max_r + 1):
            fp = false_positive_area(threshold, b, r)
            fn = false_negative_area(threshold, b, r)
            error = fp * false_positive_weight + fn * false_negative_weight
            if error < min_error:
                min_error = error
                opt = (b, r)
    return opt


# endregion

# region: Quality Control
def process_cluster(cluster: List[Any]) -> List[Any]:
    return cluster[:1]
# endregion



@register_etl
def deduplication___minhash___lsh_jaccard(
    spark,
    data: Union[RDD, DataFrame],
    threshold=0.7,
    ngram_size=5,
    min_length=5,
    num_perm=250,
    band_n=None,
    row_per_band=None,
    subset="text",
    seed=42,
    *args,
    **kwargs,
):
    """
    fuzzy deduplication using minhash and lsh
    reference - ChenghaoMou/text-dedup

    args:
        threshold: similarity threshold
        ngram_size: n-gram size
        min_length: minimum token length of document to be considered
        num_perm: number of permutations
        band_n: number of bands
        row_per_band: number of rows per band
        subset: column to deduplicate on
    """
    if isinstance(data, RDD):
        data = data.toDF()

    RNG = np.random.RandomState(seed)

    if band_n is None or row_per_band is None:
        band_n, row_per_band = optimal_param(threshold, num_perm)

    HASH_RANGES: List[Tuple[int, int]] = [
        (i * row_per_band, (i + 1) * row_per_band) for i in range(band_n)
    ]
    PERMUTATIONS: Tuple[np.ndarray, np.ndarray] = (
        RNG.randint(1, MOD_PRIME, size=(num_perm,), dtype=DTYPE),
        RNG.randint(0, MOD_PRIME, size=(num_perm,), dtype=DTYPE),
    )

    # region: Data Loading
    data = data.withColumn("__id__", F.monotonically_increasing_id()).cache()
    # endregion

    # region: MinHash
    records: RDD = data.select("__id__", subset).rdd.cache()
    buckets: RDD = (
        records.flatMap(
            lambda x: generate_hash_values(
                content=x[1],  # subset
                idx=x[0],  # __id__
                num_perm=num_perm,
                ngram_size=ngram_size,
                min_length=min_length,
                hashranges=HASH_RANGES,
                permutations=PERMUTATIONS,
            )
        )  # (band_idx, band hash value, idx)
        .groupBy(lambda x: (x[0], x[1]))  # group by (band_idx, band hash value)
        .mapValues(lambda x: [ele[2] for ele in x])  # ((band_idx, hash value), [idx, ...])
    ).cache()
    records.unpersist()
    # endregion

    # region: Connected Components
    edges: RDD = buckets.flatMap(lambda x: generate_edges(x[1])).distinct().cache()
    buckets.unpersist()
    duplicate_edges, converged, iteration = alternating_algo(edges, max_iteration=20)
    duplicate_edges.cache()
    edges.unpersist()
    # endregion

    # region: Merge Results
    if not duplicate_edges.isEmpty():
        self_edges: pyspark.RDD = duplicate_edges.values().distinct().map(lambda x: (x, x)).cache()
        all_edges: pyspark.RDD = duplicate_edges.union(self_edges).cache()
        data = data.join(
            spark.createDataFrame(all_edges, schema=["__id__", "__component__"]),
            on="__id__",
            how="left",
        ).cache()
        duplicate_edges.unpersist()
        # endregion

        self_edges.unpersist()

        # region: Quality Control: This section is hard-coded for The Stack
        cliques: RDD = all_edges.groupBy(lambda x: x[1]).cache()
        all_edges.unpersist()

        data = data.join(
            spark.createDataFrame(
                cliques.mapValues(lambda x: process_cluster(cluster=list(x))).flatMap(
                    lambda x: [(ele[0], True) for ele in x[1]]
                ),
                schema=["__id__", "__keep__"],
            ),
            on="__id__",
            how="left",
        )
        cliques.unpersist()
        data = data.filter(F.col("__component__").isNull() | F.col("__keep__")).cache()
        data = data.drop("__component__", "__keep__").cache()
    # endregion

    # drop temporary columns
    data = data.drop("__id__").cache()

    return data


"""
Interface to check & load the configurations for installation environment

awesome_config = Config.load("/path/to/ducky_awesome_config.yaml")
awesome_config = Config.load({awesome: config})
"""

import re
import boto3

from pathlib import Path
from typing import Union
from omegaconf import OmegaConf
from omegaconf import DictConfig

from dataverse.utils.setting import SystemSetting
from dataverse.utils.api import aws_s3_read
from dataverse.utils.api import aws_s3_write
from pathlib import Path



class Config:
    """
    Interface to check & load the configurations
    lightweight wrapper for OmegaConf
    """
    def __new__(cls, *args, **kwargs):
        raise NotImplementedError("Config is not allowed to be instantiated")

    @classmethod
    def load(cls, config: Union[str, dict, DictConfig, OmegaConf, Path]):
        """
        config (Union[str, dict, OmegaConf]): config for the etl
            - str or Path: (this could has several cases)
                - path to the config file
                - s3 path to the config file
                - config string
                    - this is like when you load `yaml` file with open()
                        config = yaml.load(f)
            - dict: config dict
            - OmegaConf: config object
        """
        if isinstance(config, (str, Path)):
            if isinstance(config, Path):
                config = str(config)

            # Local File
            if Path(config).is_file():
                config = OmegaConf.load(config)

            # AWS S3
            elif config.startswith(('s3://', 's3a://', 's3n://')):
                aws_s3_matched = re.match(r's3[a,n]?://([^/]+)/(.*)', config)
                if aws_s3_matched:
                    bucket, key = aws_s3_matched.groups()
                    config_content = aws_s3_read(bucket, key)
                    config = OmegaConf.create(config_content)
                else:
                    # assume it's a config string that starts with s3
                    config_str = config
                    config = OmegaConf.create(config_str)

                    # check if it's config string or not
                    # in case of config string it should create a config object
                    # if not, it will create {'config': None}
                    if config_str in config and config[config_str] is None:
                        raise ValueError(f"config {config_str} is not a valid s3 path")
            
            # String Config
            else:
                # assume it's a config string
                config_str = config
                config = OmegaConf.create(config_str)

                # same as above, check if it's config string or not
                if config_str in config and config[config_str] is None:
                    raise ValueError(f"config {config_str} is not a valid path")

        elif isinstance(config, dict):
            config = OmegaConf.create(config)
        elif isinstance(config, (OmegaConf, DictConfig)):
            pass
        else:
            raise TypeError(f"config should be str, dict, or OmegaConf but got {type(config)}")

        return config

    @classmethod
    def save(cls, config, path: Union[str, Path]):
        if path.startswith(('s3://', 's3a://', 's3n://')):
            aws_s3_matched = re.match(r's3[a,n]?://([^/]+)/(.*)', path)
            if aws_s3_matched:
                bucket, key = aws_s3_matched.groups()
                aws_s3_write(bucket, key, config)
            else:
                raise ValueError(f"config path {path} is not a valid s3 path")
        else:
            OmegaConf.save(config, Path(path))

    @classmethod
    def default(cls, emr: bool = False):
        """
        fill the missing config with default
        """
        local_dir = f"{SystemSetting().CACHE_DIR}/.cache/dataverse/tmp"

        default = OmegaConf.create({
            'spark': {
                'master': 'local[10]',
                'appname': 'default',
                'driver': {
                    'memory': '8G',
                    'maxResultSize': '2G',
                },
                'executor': {'memory': '1G'},
                'local': {'dir': local_dir},
                'ui': {'port': 4040},
            },
            'etl': [],
        })

        if emr:
            default.update({
                'emr': {
                    'id': None,
                    'working_dir': None,
                    'name': 'dataverse_emr',
                    'release': 'emr-6.15.0',
                    'idle_timeout': 3600,

                    # master (driver)
                    'master_instance': {
                        'type': None,
                    },

                    # core (data node)
                    'core_instance': {
                        'type': None,
                        'count': 2,
                    },

                    # task (executors)
                    'task_instance': {
                        'type': None,
                        'count': 0,
                    },

                    # EMR cluster created by dataverse or user
                    'auto_generated': None,

                    # iam
                    'role': {
                        'ec2': {
                            'name': None,
                            'policy_arns': None,
                        },
                        'emr': {
                            'name': None,
                            'policy_arns': None,
                        }
                    },
                    'instance_profile': {
                        'name': None,
                        'ec2_role': None,
                    },

                    # TODO: allow more options to customize e.g. cidr, tag, etc.
                    #       but make sure vpc is temporary and not shared
                    'vpc': {
                        'id': None,
                    },
                    'subnet': {
                        'id': None,
                        'public_id': None,
                        'private_id': None,
                        'public': True,
                    },
                    'security_group': {
                        'id': None,
                    },
                    'gateway': {
                        'id': None,
                    },
                    'route_table': {
                        'id': None,
                    },
                    'elastic_ip': {
                        'id': None,
                    },
                    'nat_gateway': {
                        'id': None,
                    },
                }
            })

        return default

    @classmethod
    def set_default(cls, config: OmegaConf, emr: bool = False):
        """
        set the missing config args with default
        """
        return OmegaConf.merge(cls.default(emr=emr), config)





from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from dataverse.etl.registry import register_etl

import re
import random
from typing import Union



@register_etl
def pii___card___replace_card_number(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    pattern=r'(\d{4}-\d{4}-\d{4}-\d{4})',
    random_pii=True,
    replace_pii=False,
    replace_token='[CARD_NUMBER]',
    start_token='',
    end_token='',
    *args,
    **kwargs
):
    """
    Replace card number with a random number or a token

    `replace_pii` takes precedence over `random_pii`
    - e.g when both are True, the card number will be replaced with the token
    - e.g. this is 1234-1234-1234-1234 -> this is [CARD_NUMBER]

    `start_token` and `end_token` are used to append the token to the start and end of the card number
    - it doens't matter with `random_card_number` or `replace_card_number` is True or False

    [ example ]
    <input>
        - text = 'card number is 1234-1234-1234-1234.'

    <output>
        - random pii
            - text = 'card number is 2238-1534-1294-1274.'
        - replace pii
            - replace_token = '[CARD_NUMBER]'
            - text = 'card number is [CARD_NUMBER].'
        - start token
            - start_token = '[CARD_NUMBER_START]'
            - text = 'card number is [CARD_NUMBER_START]1234-1234-1234-1234.'
        - end token
            - end_token = '[CARD_NUMBER_END]'
            - text = 'card number is 1234-1234-1234-1234[CARD_NUMBER_END].'

    args:
        subset (str): subset or columns to consider
        pattern (str): regex pattern to find
        random_pii (bool): if True, replace the pii with a random number
        replace_pii (bool): if True, replace the pii with the `replace_token`
        replace_token (str): token to replace the pii with
        start_token (str): start token to append where the pattern is found
        end_token (str): end token to append where the pattern is found
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    def _replace_match(match):
        match = match.group()
        if replace_pii:
            match = replace_token
        elif random_pii:
            match = re.sub('\d', lambda x: str(random.randint(0, 9)), match)

        return f"{start_token}{match}{end_token}"

    def _replace_pii(row):
        row[subset] = re.sub(pattern, _replace_match, row[subset])
        return row

    data = data.map(_replace_pii)

    return data




"""
special purpose to create fake data for testing or debugging
"""

from dataverse.etl import register_etl
from dataverse.utils.format import huggingface2parquet
from dataverse.utils.format import load_huggingface_dataset

import json
from faker import Faker
from typing import Union, List


@register_etl
def data_ingestion___test___generate_fake_ufl(
    spark,
    n=100,
    repartition=20,
    verbose=True,
    *args,
    **kwargs
):
    """
    generate fake data for testing or debugging

    Args:
        spark (SparkSession): spark session
        n (int): the number of data to generate
        repartition (int): the number of partitions
        verbose (bool): whether to print the information of the dataset
    """
    faker = Faker()
    def _generate_fake_ufl(n=100):
        while n > 0:
            n -= 1
            yield {
                "id": faker.uuid4(),
                "name": 'test_fake_ufl',
                "text": faker.text(),
                "meta": json.dumps({
                    "name": faker.name(),
                    "age": faker.random_int(0, 100),
                    "address": faker.address(),
                    "job": faker.job(),
                }),
            }
    rdd = spark.sparkContext.parallelize(_generate_fake_ufl(n=n))
    rdd = rdd.repartition(repartition)

    return rdd





from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from dataverse.etl.registry import register_etl

import re
import unicodedata
from typing import Union


@register_etl
def cleaning___document___split_by_word(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    word_per_chunk=100,
    delimiter=" ",
    *args,
    **kwargs
):
    """
    split documents to smaller chunks by word

    [ example ]
    - word_per_chunk = 2
    - delimiter = " "
    - text = "hello world, how are you?"
    - result = ["hello world,", "how are", "you?"]

    [ caveat ]
    - NO normalization is done here!
        - this doesn't consider the whitespace normalization
        - recommend to use other normalization before this
    - All the keys from the original row are copied to all the new rows created
        - `id` is not unique anymore
        - make sure `id` is assigned after this step

    args:
        subset (str): subset or columns to consider
        word_per_chunk (int): number of words per chunk
        delimiter (str): delimiter to split the text
            - default: " "
            - decide the unit of word
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    def _split_by_word(row):
        words = row[subset].split(delimiter)

        # Create chunks
        chunks = []
        for i in range(0, len(words), word_per_chunk):
            chunks.append(delimiter.join(words[i: i+word_per_chunk]))

        # Create a new dictionary for each chunk with all the keys from the original row
        return [{**row, subset: chunk} for chunk in chunks]

    data = data.flatMap(_split_by_word)

    return data




import pkg_resources


def pip_get_package_path(package_name):
    try:
        package = pkg_resources.get_distribution(package_name)
        return package.location
    except pkg_resources.DistributionNotFound:
        print(f"Package '{package_name}' is not installed.")
        return None





from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from dataverse.etl import BaseETL
from dataverse.etl import register_etl
from dataverse.etl import ETLRegistry
from dataverse.etl.registry import ETLStructure

from typing import Union


@register_etl
def __sample___github___using_decorator(data: Union[RDD, DataFrame], *args, **kwargs):
    """
    decorator will convert this function to BaseETL class
    """
    print("sample using decorator")
    return data

@register_etl
def __sample___github___config(data: Union[RDD, DataFrame], config: dict = None, *args, **kwargs):
    """
    decorator will convert this function to BaseETL class
    """
    print("config says", config)
    return data

if __name__ == "__main__":
    registry = ETLRegistry()

    print("[ Testing ] registry etl using decorator")
    # this could seem like a function but it is actually a BaseETL class
    etl = __sample___github___using_decorator
    etl()(data=None)
    print("is subclass of ETLStructure?", issubclass(etl, ETLStructure), "\n")

    print("[ Testing ] registry etl using decorator with config")
    etl = __sample___github___config
    etl()(data=None, config={"hello": "world"})
    print("is subclass of ETLStructure?", issubclass(etl, ETLStructure), "\n")

    # check is it properly registryed
    print("[ Testing ] check is it properly registry")
    print("="*50)
    print(registry._registry)
    print("="*50)



import ast

def python_is_script_executable(file_path, verbose=False):
    """
    check if a python script is executable
    in other words, check if the python script does not contains any declaration nodes
    (imports, functions, classes, etc.)

    declaration nodes:
    - imports
    - functions
    - classes
    - variables

    Args:
        file_path (str): path to the python script to check
    Returns:
        bool: True if the python script is executable, False otherwise
    """
    with open(file_path, 'r') as file:
        source_code = file.read()

    # Parse source code into an AST
    module = ast.parse(source_code)
    for node in module.body:
        if not isinstance(node, (
            ast.Import,
            ast.ImportFrom,
            ast.FunctionDef,
            ast.ClassDef,
            ast.Assign
        )):
            if verbose:
                print("found executable code: {}".format(node))
            return True

    if verbose:
        print("found no executable code")
    return False





from pyspark.rdd import RDD
from pyspark.sql import DataFrame
from pyspark.sql.functions import split, posexplode
from pyspark.sql.functions import collect_list
from pyspark.sql import functions as F

from dataverse.etl.registry import register_etl

import re
import functools
from typing import Union


def filter_lines(row, subset='text'):
    row = row.asDict()
    text = row[subset]
    line_ids = row['line_ids']

    text_lines = text.split('\n')
    filtered_texts = "\n".join([text_lines[line_i] for line_i in sorted(line_ids)])

    del row['line_ids']
    row[subset] = filtered_texts

    return row


@register_etl
def deduplication___common_crawl___exact_line(spark, data: Union[RDD, DataFrame], subset='text', *args, **kwargs):
    """
    exact line deduplication - which is a line by line deduplication.

    strip & lower is applied to the line text before deduplication
    but this will not be applied to the original text

    - input
        - 'DuckY'
        - 'dUKCY '
    - after normalization
        - 'ducky'
        - 'ducky'
    - after deduplication
        - 'ducky'
    - output
        - 'DuckY'


    args:
        subset (str): subset or columns to consider if duplicated
    """
    if isinstance(data, RDD):
        data = data.toDF()

    data = data.cache()
    data = data.withColumn("__id__", F.monotonically_increasing_id())

    assert isinstance(data, DataFrame), f"data must be DataFrame, got {type(data)}"
    line_data = data.select('__id__', posexplode(split(data[subset], '\n')).alias('line_id', 'line'))
    line_data = line_data.withColumn('line', F.lower(F.trim(line_data['line'])))
    line_data = line_data.dropDuplicates(subset=['line'])
    line_data = line_data.groupBy('__id__').agg(collect_list('line_id').alias('line_ids'))

    merged_data = data.join(line_data, on=['__id__'], how='inner')
    data.unpersist()
    line_data.unpersist()

    # remove __id__
    merged_data = merged_data.drop('__id__')

    # filter the lines using the line_ids
    merged_data = merged_data.rdd.map(functools.partial(filter_lines, subset=subset))

    return merged_data



from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from dataverse.etl import register_etl
from dataverse.utils.format import huggingface2parquet
from dataverse.utils.format import load_huggingface_dataset

from typing import Union, List
from omegaconf import ListConfig

import datasets


"""
Supported datasets:
https://huggingface.co/datasets/cerebras/SlimPajama-627B
"""


@register_etl
def data_ingestion___slim_pajama___parquet2ufl(spark, input_paths, repartition=20, *args, **kwargs):
    """
    convert parquet file to ufl
    """
    df = spark.read.parquet(*input_paths)
    rdd = df.rdd.repartition(repartition)
    rdd = rdd.map(lambda row: row.asDict())
    return rdd

@register_etl
def data_ingestion___slim_pajama___hf2ufl(
    spark,
    name_or_path : Union[str, List[str]] = 'cerebras/SlimPajama-627B',
    split=None,
    from_disk=False,
    repartition=20,
    verbose=True,
    *args,
    **kwargs
):
    """
    convert huggingface dataset to ufl

    Args:
        spark (SparkSession): spark session
        name_or_path (str or list): the name or path of the huggingface dataset
        split (str): the split of the dataset
        from_disk (bool): whether to load from disk
            - no split is allowed when from_disk is True
        repartition (int): the number of partitions
        verbose (bool): whether to print the information of the dataset
    """
    dataset = load_huggingface_dataset(name_or_path, split=split, from_disk=from_disk)
    parquet_path = huggingface2parquet(dataset, verbose=verbose)

    df = spark.read.parquet(parquet_path)
    rdd = df.rdd.repartition(repartition)
    rdd = rdd.map(lambda row: row.asDict())
    return rdd





from pyspark.rdd import RDD
from pyspark.sql import DataFrame
from pyspark.sql import functions as F

from typing import Union

from dataverse.etl.registry import register_etl



@register_etl
def cleaning___table___merge_col_vertical(
    spark,
    data: Union[RDD, DataFrame],
    col1=None,
    col2=None,
    merge_col_name="merge_col",
    *args,
    **kwargs
):
    """
    merge 2 column vertically to 1 column
    what does this mean? here is example

    [ before ]
    -----------------------
    col1 col2 species
    1    2    duck
    3    4    duck
    5    6    ducky
    -----------------------

    cleaning___table___merge_col_vertical(
        ...,
        col1="col1",
        col2="col2",
        merge_col_name="number",
        ...
    )

    [ after ]
    -----------------------
    number species
    1      duck
    2      duck
    3      duck
    4      duck
    5      ducky
    6      ducky
    -----------------------

    args:
        col1: column to merge
        col2: column to merge
        merge_col_name: name of merged column
    """
    if isinstance(data, RDD):
        data = data.toDF()

    assert col1 is not None, "col1 must be specified"
    assert col2 is not None, "col2 must be specified"

    rest_cols = [c for c in data.columns if c not in [col1, col2]]
    df1 = data.select(*rest_cols, F.col(col1).alias(merge_col_name))
    df2 = data.select(*rest_cols, F.col(col2).alias(merge_col_name))

    # union the dataframes
    data = df1.union(df2)

    return data





"""
Data loading to Parquets
"""

import os
from pyspark.rdd import RDD
from pyspark.sql import DataFrame
from dataverse.etl import register_etl


@register_etl
def data_load___parquet___ufl2parquet(spark, ufl, save_path, repartition=1, *args, **kwargs):
    """
    save data to parquet and return the path
    """
    if os.path.exists(save_path):
        raise ValueError(f"save_path {save_path} already exists")

    if isinstance(ufl, RDD):
        ufl = ufl.toDF()

    assert isinstance(ufl, DataFrame), f"ufl must be RDD or DataFrame, got {type(ufl)}"

    ufl = ufl.repartition(repartition)
    ufl.write.parquet(save_path, mode="overwrite")

    return save_path




"""
Code is from EleutherAI/dps
https://github.com/EleutherAI/dps/blob/master/dps/spark/jobs/dedup_job.py

This is a migration of the deduplication job from the DPS project to the Dataverse.
"""

import random
import binascii
import numpy as np
from itertools import combinations

from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from dataverse.etl import register_etl

from typing import Union, List


MERSENNE_PRIME = (1 << 61) - 1
MAX_HASH = (1 << 32) - 1
HASH_RANGE = 1 << 32


def shingle_word(text: str, n_gram: int = 15, char_level: bool = False) -> List[str]:
    """
    example
    -------
    >>> shingle_word("hello world from ducky", n_gram=2)
    ['hello_world', 'world_from', 'from_ducky']

    >>> shingle_word("hello world from ducky", n_gram=2, char_level=True)
    ['h_e', 'e_l', 'l_l', 'l_o', 'o_w', 'w_o', 'o_r', 'r_l', 'l_d', 'd_f', 'f_r', 'r_o', 'o_m', 'm_d', 'd_u', 'u_c', 'c_k', 'k_y']
    """
    res = []
    text_words = text.split() if not char_level else text

    for i in range(len(text_words)):
        shingle = text_words[i : i + n_gram]

        if len(shingle) == n_gram:
            res.append("_".join(shingle).encode("utf-8"))

    return res


def generate_minhash(shingles: List, num_perm: int = 64, seed: int = 1) -> np.array:
    def hashfunc(b: bytes) -> bytes:
        return binascii.crc32(b) & MAX_HASH

    hashvalues = np.ones(num_perm, dtype=np.uint64) * MAX_HASH

    generator = np.random.RandomState(seed)
    permutations = np.array(
        [
            (
                generator.randint(1, MERSENNE_PRIME, dtype=np.uint64),
                generator.randint(0, MERSENNE_PRIME, dtype=np.uint64),
            )
            for _ in range(num_perm)
        ],
        dtype=np.uint64,
    ).T

    for shingle in shingles:
        hv = hashfunc(shingle)
        a, b = permutations
        phv = np.bitwise_and((a * hv + b) % MERSENNE_PRIME, np.uint64(MAX_HASH))
        hashvalues = np.minimum(phv, hashvalues)

    return hashvalues


def jaccard_by_hashvalues(src_hashvalues, tgt_hashvalues) -> float:
    if len(src_hashvalues) != len(tgt_hashvalues):
        raise ValueError()

    return np.float(np.count_nonzero(src_hashvalues == tgt_hashvalues)) / np.float(
        len(src_hashvalues)
    )


def expand_instances_by_minhash(
    data, expand_size: int, n_gram: int, seed: int = 1, char_level: bool = False
):
    shingles = shingle_word(data["text"], n_gram=n_gram, char_level=char_level)
    minhashes = generate_minhash(shingles, num_perm=expand_size, seed=seed)

    for mh in minhashes.tolist():
        yield (str(mh), [dict(**data, shingles=shingles, hashvalues=minhashes)])


def explore_dedup_instance(hash_groups, threshold: float = 0.8):
    if len(hash_groups) <= 1:
        return

    group_represent_text = hash_groups[0][
        "text"
    ]  # not to remove all text instances in group.
    pairs = combinations(hash_groups, 2)

    for d_1, d_2 in pairs:
        sim_score = jaccard_by_hashvalues(d_1["hashvalues"], d_2["hashvalues"])
        if sim_score >= threshold:
            dedup_text = [d_1["text"], d_2["text"]]
            if group_represent_text in dedup_text:
                yield dedup_text[0] if dedup_text[
                    0
                ] != group_represent_text else dedup_text[1]
            else:
                yield random.choice(dedup_text)


@register_etl
def deduplication___polyglot___minhash(
    spark,
    data: Union[RDD, DataFrame],
    expand_size: int = 64,
    n_gram: int = 15,
    seed: int = 1,
    char_level: bool = False,
    sim_threshold: float = 0.8,
    *args,
    **kwargs,
):
    """
    fuzzy deduplication

    for UFL
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    overlap_kv_rdd: RDD = (
        data.flatMap(
            lambda x: expand_instances_by_minhash(
                x,
                expand_size=expand_size,
                n_gram=n_gram,
                seed=seed,
                char_level=char_level,
            )
        )
        .reduceByKey(lambda x, y: x + y)
        .flatMap(
            lambda x: explore_dedup_instance(x[1], threshold=sim_threshold)
        )
        .distinct()
        .map(lambda x: (x, dict(text=x)))
        .cache()
    )

    data = data.map(lambda x: (x["text"], x)).subtractByKey(overlap_kv_rdd).map(
        lambda x: x[1]
    )
    return data





from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from typing import Union, List, Tuple
from operator import add

from dataverse.etl import register_etl


@register_etl
def utils___statistics___korean_nouns(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    *args,
    **kwargs
) -> RDD[List[Tuple[str, int]]]:
    """
    get the frequency of each noun in the given subset of the data

    [ example ]
    Input:
        [
            {'text': '오리는 꽥꽥 웁니다. 거위는'},
            {'text': '안녕 세상!'},
            {'text': '사람들은 꽥꽥 울지 않습니다. 오리가 웁니다'},
        ]
    Output:
        [('오리', 2), ('거위', 1), ('세상', 1), ('사람', 1)]

    caveat:
        - this works for `KOREAN` only
        - this is not returning the ufl rather the frequency of each noun

    Args:
        spark (SparkSession): SparkSession
        data (Union[RDD, DataFrame]): data to extract the nouns from
        subset (str): subset of the data to extract the nouns from
    """

    # konlpy & mecab
    try:
        from konlpy.tag import Mecab
    except ImportError:
        raise ImportError(
            'Please install konlpy & Mecab:\n'
            'pip install konlpy\n'
            'pip install mecab-python3\n'
        )

    if isinstance(data, DataFrame):
        data = data.rdd

    mecab = Mecab()
    def _parse_korean_nouns(text):
        try:
            if text is not None:
                return mecab.nouns(text)
            else:
                return []
        except Exception as e:
            # Log the exception for debugging purposes
            return []

    # Count the frequency of each noun
    data = data.flatMap(lambda x: _parse_korean_nouns(x[subset]))
    noun_counts = data.map(lambda noun: (noun, 1)).reduceByKey(add)

    return noun_counts




from pyspark.rdd import RDD
from pyspark.sql import DataFrame
from pyspark.sql.functions import split, posexplode
from pyspark.sql.functions import collect_list

from dataverse.etl.registry import register_etl

import re
import unicodedata
from typing import Union



@register_etl
def cleaning___number___normalize(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    assign_number=0,
    *args,
    **kwargs
):
    """
    convert all the number to assigned number (e.g. 0)
    example
    - 1234 -> 0000
    - 1234.5678 -> 0000.0000

    Code is from facebookresearch/cc_net
    https://github.com/facebookresearch/cc_net/blob/main/cc_net/text_normalizer.py

    args:
        subset (str): subset or columns to consider
        assign_number (int): number to assign
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    def _normalize_number(row):
        row[subset] = re.sub(r"\d", str(assign_number), row[subset])
        return row

    # assign_number is between 0 ~ 9
    assert assign_number in range(10), f"assign_number should be between 0 ~ 9 but got {assign_number}"
    data = data.map(_normalize_number)

    return data





from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from dataverse.etl.registry import register_etl

from typing import Union


@register_etl
def deduplication___exact___column(spark, data: Union[RDD, DataFrame], subset=['text'], *args, **kwargs):
    """
    exact column deduplication

    args:
        subset: subset or columns to consider if duplicated
    """
    if isinstance(data, RDD):
        data = data.toDF()

    assert isinstance(data, DataFrame), f"data must be DataFrame, got {type(data)}"
    data = data.dropDuplicates(subset=subset)
    return data



import os
import datasets
from pathlib import Path
from omegaconf import ListConfig

from dataverse.utils.setting import SystemSetting


def load_huggingface_dataset(name_or_path, split=None, from_disk=False):
    """
    load huggingface dataset
    
    Args:
        name_or_path (str or list): the name or path of the huggingface dataset
        split (str): the split of the dataset
    """
    if from_disk:
        if split is not None:
            raise ValueError("split is not supported when from_disk is True")

        # load huggingface dataset
        if isinstance(name_or_path, str):
            dataset = datasets.load_from_disk(name_or_path)
        elif isinstance(name_or_path, list):
            dataset = datasets.load_from_disk(*name_or_path)
        elif isinstance(name_or_path, ListConfig):
            dataset = datasets.load_from_disk(*name_or_path)
        else:
            raise ValueError(f"Unsupported type of name_or_path: {type(name_or_path)}")
    else:
        # load huggingface dataset
        if isinstance(name_or_path, str):
            dataset = datasets.load_dataset(name_or_path, split=split)
        elif isinstance(name_or_path, list):
            dataset = datasets.load_dataset(*name_or_path, split=split)
        elif isinstance(name_or_path, ListConfig):
            dataset = datasets.load_dataset(*name_or_path, split=split)
        else:
            raise ValueError(f"Unsupported type of name_or_path: {type(name_or_path)}")

    return dataset


def huggingface2parquet(
        dataset: datasets.Dataset,
        cache_dir: str = None,
        verbose: bool = True,
        **kwargs
    ):
    """
    Convert a huggingface dataset to parquet format and save it to the path.
    
    Args:
        dataset (datasets.Dataset): a huggingface dataset
        cache_dir (str): cache path to save the dataset
        verbose (bool): whether to print the information of the dataset
    """
    # check the dataset which has train, test, validation or other splits
    # concatenate all the splits into one
    dataset_list = []

    # check the dataset has splits
    try:
        for split in dataset.keys():
            dataset_list.append(dataset[split])
    except:
        dataset_list.append(dataset)

    dataset = datasets.concatenate_datasets(dataset_list)

    # save the dataset to parquet
    # FIXME: this is a temporary solution to store the dataset in the package root path
    #        we will change it to a better solution in the future
    if cache_dir is None:
        # save the parquet at package root path
        cache_dir = SystemSetting().CACHE_DIR

    dataset_path = f"{cache_dir}/.cache/dataverse/dataset/huggingface_{dataset._fingerprint}.parquet"

    # check the dataset exist
    if os.path.exists(dataset_path):
        if verbose:
            print(f"Dataset already exists at {dataset_path}")
        return dataset_path

    os.makedirs(f"{cache_dir}/.cache/dataverse/dataset", exist_ok=True)
    dataset.to_parquet(dataset_path)

    return dataset_path

if __name__ == "__main__":
    # test the function
    dataset = load_huggingface_dataset(["glue", "mrpc"])
    dataset_path = huggingface2parquet(dataset, verbose=True)

    print(f"Dataset saved at {dataset_path}")




from pyspark.rdd import RDD
from pyspark.sql import DataFrame
from pyspark.sql import functions as F

from typing import Union, Optional

from dataverse.etl.registry import register_etl



@register_etl
def cleaning___length___char_len_filter(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    min_len: int = None,
    max_len: int = None,
    *args,
    **kwargs
):
    """
    filter by char length

    min_len <= len <= max_len
    - if min_len is None, then len <= max_len
    - if max_len is None, then len >= min_len

    args:
        subset: column to filter
        min_len: minimum length to filter
        max_len: maximum length to filter

    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    assert min_len is not None or max_len is not None, "min_len and max_len cannot be None at the same time"

    if min_len is not None and max_len is not None:
        data = data.filter(lambda row: min_len <= len(row[subset]) <= max_len)
    elif min_len is None:
        data = data.filter(lambda row: len(row[subset]) <= max_len)
    elif max_len is None:
        data = data.filter(lambda row: min_len <= len(row[subset]))

    return data


@register_etl
def cleaning___length___word_len_filter(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    min_len: int = None,
    max_len: int = None,
    *args,
    **kwargs
):
    """
    filter by word length

    min_len <= len <= max_len
    - if min_len is None, then len <= max_len
    - if max_len is None, then len >= min_len

    args:
        subset: column to filter
        min_len: minimum length to filter
        max_len: maximum length to filter

    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    assert min_len is not None or max_len is not None, "min_len and max_len cannot be None at the same time"

    if min_len is not None and max_len is not None:
        data = data.filter(lambda row: min_len <= len(row[subset].split()) <= max_len)
    elif min_len is None:
        data = data.filter(lambda row: len(row[subset].split()) <= max_len)
    elif max_len is None:
        data = data.filter(lambda row: min_len <= len(row[subset].split()))

    return data




from pyspark.rdd import RDD
from pyspark.sql import DataFrame

from dataverse.etl.registry import register_etl

import html2text

from typing import Union


@register_etl
def cleaning___html___extract_plain_text(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    use_trafilatura=False,
    *args,
    **kwargs
):
    """
    extract plain text from html

    [ caveat ]
    - `html2text` addes double newline after each paragraph
        - we are not handling it at this point
    - why I've put `trafilatura` as an option?
        - because the extracting plain text with `trafilatura` does not seems to work well
            - [ OK ] case
                ```python
                text = "<body><h1>My First Heading</h1><p>My first paragraph.</p></body>"

                # html2text
                print(html2text.html2text(text))
                >>> '# My First Heading\n\nMy first paragraph.\n\n'

                # trafilatura
                print(trafilatura.html2txt(text))
                >>> 'My First HeadingMy first paragraph.'
                ```
            - [ ERROR ] case (trafilatura removes all the text)
                ```python
                text = "<p>hello <br> nice to meet you.</p>"

                # html2text
                print(html2text.html2text(text))
                >>> 'hello  \nnice to meet you.\n\n'

                # trafilatura
                print(trafilatura.html2txt(text))
                >>> ''
                ```

    args:
        subset (str): subset or columns to consider
        use_trafilatura (bool): use trafilatura instead of html2text
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    # this is optional
    if use_trafilatura:
        import trafilatura
        def _html2txt(row):
            row[subset] = trafilatura.html2txt(row[subset])
            return row
    else:
        def _html2txt(row):
            row[subset] = html2text.html2text(row[subset])
            return row

    data = data.map(_html2txt)

    return data




"""
Interface for system setting
"""

import os
import re
import uuid
import json
import boto3
import pyspark
from pathlib import Path

import dataverse
from dataverse.utils.api import aws_check_credentials
from dataverse.utils.api import aws_s3_create_bucket
from dataverse.utils.api import aws_s3_list_buckets


class SystemSetting:
    """
    System Setting CRUD interface

    system setting holds all the variables
    that influence the behavior of the dataverse system.

    Also, this class is a singleton class, so you can use it anywhere in the code
    
    [ MEMORY ONLY ]
    - system setting is stored in memory only
    - system setting is not persistent

    [ Update by Env Variable ]
    - system setting can be updated by env variable

    [ Manual Update ]
    - default system setting can be updated manually
    - check `default_setting()`

    [ No Update after Initialization ]
    - system could be updated but not reflected in the program
    - need to restart the program to use new system setting
    """
    # Singleton
    _initialized = False

    # TODO: system setting per user [Candidate]
    ...

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(SystemSetting, cls).__new__(cls)
        return cls.instance

    def __init__(self):
        # when the class is initialized, this is called everytime
        # regardless of the singleton. So adding the flag to check
        if self._initialized:
            return

        self.default_setting()
        self.update_by_env()
        self._initialized = True

    def _get_aws_bucket(self, verbose=True):
        """
        the bucket will be used to store the dataverse info
        - cache
        - log
        - etc

        - format
            - dataverse-{MAGIC_NUMBER}-{UUID}
        """
        # if aws credential is not valid, return None
        if not aws_check_credentials():
            return None

        identify_prefix = f'dataverse-{self.MAGIC_NUMBER}-'
        for bucket in aws_s3_list_buckets():
            if identify_prefix in bucket:

                # check if the last part is uuid
                uuid_part = bucket.replace(identify_prefix, "")
                try:
                    uuid.UUID(uuid_part)

                    # Use this bucket for your package operations
                    if verbose:
                        print("Detected Dataverse Bucket: " + bucket)

                    return bucket
                except ValueError:
                    # not a valid UUID, so ignore this bucket
                    pass 

        # if there is no relevant bucket, create one
        bucket = f'dataverse-{self.MAGIC_NUMBER}-{uuid.uuid1()}'
        aws_s3_create_bucket(bucket)

        return bucket


    def default_setting(self):
        """
        Reset the system setting to default

        Default setting:
        - `MAGIC_NUMBER`: magic number for dataverse
        - `CACHE_DIR`: default cache directory
        - `IS_CLI`: if the program is running in CLI mode
        - `AWS_BUCKET`: default aws bucket name for dataverse info
        - `SPARK_VERSION`: spark version
        - `HADOOP_VERSION`: hadoop version
        """
        self.system_setting = {}

        # MAGIC NUMBER
        # dv - Dataverse
        # 42 - (The Hitchhiker's Guide to the Galaxy)
        #      The Answer to the Ultimate Question of Life, the Universe, and Everything
        self.MAGIC_NUMBER = "dv42"

        # DATAVERSE
        self.DATAVERSE_HOME = os.path.dirname(dataverse.__file__)

        # HARD CODED DEFAULT SETTING
        self.CACHE_DIR = Path.home().as_posix()
        self.IS_CLI = False

        # AWS SETTING
        self.AWS_BUCKET = self._get_aws_bucket()

        # SPARK VERSION
        self.SPARK_VERSION = pyspark.__version__

        # HADOOP VERSION
        jars = Path(pyspark.__file__).parent / "jars"
        hadoop_jar = list(jars.glob("hadoop-client-runtime*.jar"))
        self.HADOOP_VERSION = re.findall(r"\d+\.\d+\.\d+", hadoop_jar[0].name)[-1]

        # TODO: add more default setting here
        ...

    def update_by_env(self):
        """
        Update the system setting by env variable
        """
        # check if the env variable is set
        for key in self.system_setting:
            if key in os.environ:
                self.system_setting[key] = os.environ[key]

    def check_naming_convention(self, key):
        """
        1. only CAPITALIZED format
            - e.g. CACHE_DIR (O)
            - e.g. cache_dir (X)
        2. only alphanumeric and underscore
            - e.g. CACHE_DIR2 (O)
            - e.g. cache-dir (X)
            - e.g. CACHE_@DIR (X)
        3. only one underscore between words
            - e.g. CACHE__DIR (X)
        4. no underscore at the start/end of the key
            - e.g. _CACHE_DIR (X)
            - e.g. CACHE_DIR_ (X)
        """
        # 1. only CAPITALIZED format
        if key != key.upper():
            raise ValueError(f"key [ {key} ] is not in Capitalized format")

        # 2. only alphanumeric and underscore
        for char in key:
            if not char.isalnum() and char != "_":
                raise ValueError(f"key [ {key} ] should only contains alphanumeric and underscore")

        # 3. only one underscore between words
        if "_" in key:
            # check if there is only one underscore
            divided_keys = key.split("_")
            if "" in divided_keys:
                raise ValueError(f"key [ {key} ] contains more than one underscore")

        # 4. no underscore at the start/end of the key
        if key.startswith("_") or key.endswith("_"):
            raise ValueError(f"key [ {key} ] contains underscore at the start/end of the key")

    def get(self, key):
        """
        """
        if key not in self.system_setting:
            raise KeyError(f"key [ {key} ] does not exist in SYSTEM setting")
        return self.system_setting[key]

    def set(self, key, value):
        """
        """
        self.check_naming_convention(key)
        self.system_setting[key] = value

    # Support dot-like access, e.g. setting.CACHE_DIR
    def __getattr__(self, key):
        if key in [
            "_initialized",
            "system_setting"
        ]:
            return super().__getattr__(key)
        else:
            return self.get(key)

    def __setattr__(self, key, value):
        if key in [
            "_initialized",
            "system_setting"
        ]:
            super().__setattr__(key, value)
        else:
            self.set(key, value)

    # Support dict-like access, e.g. setting["CACHE_DIR"]
    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        self.set(key, value)

    def delete(self, key):
        """
        """
        if key in self.system_setting:
            self.system_setting.pop(key, None)
        else:
            raise KeyError(f"key [ {key} ] does not exist in SYSTEM setting")

    def list(self):
        """
        List all settings
        """
        print(self.system_setting)

    def __repr__(self):
        return json.dumps(self.system_setting, indent=4)

    def __str__(self):
        return json.dumps(self.system_setting, indent=4)




"""
Load Arrow

support direct loading of arrow saved huggingface dataset to spark dataframe
"""

from dataverse.etl import register_etl

import os
import glob
import pyarrow as pa
import numpy as np
from typing import Union, List
from omegaconf import ListConfig



def find_arrow_paths(directory):
    """find *.arrow files recursively"""
    if isinstance(directory, str):
        return glob.glob(os.path.join(directory, '**/*.arrow'), recursive=True)
    elif isinstance(directory, list) or isinstance(directory, ListConfig):
        arrow_paths = []
        for d in directory:
            arrow_paths.extend(find_arrow_paths(d))
        return arrow_paths

    raise ValueError(f"directory must be str or list, got {type(directory)}")

def get_dir_size(arrow_paths):
    total_size = 0
    for fp in arrow_paths:
        # skip if it is not `.arrow` file
        if not fp.endswith('.arrow'):
            continue

        # skip if it is symbolic link
        if not os.path.islink(fp):
            total_size += os.path.getsize(fp)

    return total_size


def arrow_table_to_dict(arrow_path):
    """
    speed 10000 take - 70ms

    faster than
    - pyarrow -> pydict direct loading
    - pyarrow -> pandas -> pydict loading

    TODO: speed and memory improvement
    """
    in_memory_stream = pa.input_stream(arrow_path)
    opened_stream = pa.ipc.open_stream(in_memory_stream)
    table = opened_stream.read_all()

    # get schema for field names
    schema = table.schema

    rows = []
    # iterate over each row
    for row in range(table.num_rows):
        row_data = {schema.field(col).name: table.column(col)[row].as_py() for col in range(table.num_columns)}
        rows.append(row_data)

    return rows


@register_etl
def data_ingestion___arrow___hf2raw(
    spark,
    path : Union[str, List[str]],
    sample_n=-1,
    arrow_partition_mb_size=-1,
    raw_partition_mb_size=256,
    repartition=-1,
    seed=42,
    verbose=True,
    *args,
    **kwargs
):
    """
    direct loading of arrow saved huggingface dataset to raw format as dict

    [ usage ]
    >>> import datasets
    >>> dataset = datasets.load_dataset('ducky')
    >>> dataset.save_to_disk('your/path/to/ducky')
    >>> data_ingestion___arrow___hf2raw()(spark, 'your/path/to/ducky')

    [ default setting ]
    - arrow paths are repartitioned by the number of arrow files
    - raw data is repartitioned by the size of each partition (256MB)
        - if you want to manually choose the number of partitions, set `repartition`

    Args:
        spark (SparkSession): spark session
        path (str or list): the path of the arrow folders
        sample_n (int): the number of arrow files to be sampled
            - if sample_n is -1, all arrow files will be loaded
        arrow_partition_mb_size (int): the size of each arrow partition in MB
            - if arrow_partition_size is -1, it will repartition arrow files by the number of arrow files
            - this assumes that arrow file size is evenly distributed
                - when there is data skew in arrow file size recommend default (-1)
        raw_partition_mb_size (int): the size of each raw partition in MB
            - activated only when repartition is -1
        seed (int): the seed for sampling
        repartition (int): manually choose the number of partitions
        verbose (bool): whether to print the information of the dataset
    """
    arrow_paths = find_arrow_paths(path)
    assert len(arrow_paths) > 0, f"no arrow files found in {path}"

    # sample from the arrow files
    if sample_n > 0 and sample_n < len(arrow_paths):
        np.random.seed(seed)
        arrow_paths = np.random.choice(arrow_paths, size=sample_n, replace=False)

    if arrow_partition_mb_size == -1:
        # if data is skewed, recommend to use default (-1)
        arrow_repartition = len(arrow_paths)
    else:
        # this assume that arrow file size is evenly distributed
        assert arrow_partition_mb_size > 0, f"arrow_partition_mb_size must be positive, got {arrow_partition_mb_size}"
        arrow_total_mb_size = get_dir_size(arrow_paths) / 1024 / 1024
        arrow_repartition = arrow_total_mb_size // arrow_partition_mb_size
        arrow_repartition += 1 if arrow_total_mb_size % arrow_partition_mb_size else 0
        arrow_repartition = min(int(arrow_repartition), len(arrow_paths))

    rdd = spark.sparkContext.parallelize(arrow_paths)
    rdd = rdd.repartition(arrow_repartition)
    rdd = rdd.flatMap(arrow_table_to_dict)

    if repartition != -1:
        raw_repartition = repartition
    else:
        assert raw_partition_mb_size > 0, f"raw_partition_mb_size must be positive, got {raw_partition_mb_size}"

        arrow_total_mb_size = get_dir_size(arrow_paths) / 1024 / 1024
        raw_repartition = arrow_total_mb_size // raw_partition_mb_size
        raw_repartition += 1 if arrow_total_mb_size % raw_partition_mb_size else 0

        # count the number of data points (this is expensive)
        # this is to prevent the case where the number of data points is less than raw_repartition
        total_data_n = rdd.count()
        raw_repartition = min(int(raw_repartition), total_data_n)

    rdd = rdd.repartition(raw_repartition)

    return rdd 



import os

from setuptools import find_packages, setup

basedir = os.path.abspath(os.path.dirname(__file__))
requirements_path = os.path.join(basedir, "requirements.txt")


def get_requirements():
    """Get package requirements from a requirements file (ex: requirements.txt)."""
    with open(requirements_path, "r") as f:
        return f.read().splitlines()


def get_extras_require():
    extras_require = {
        "aws": [
            "awscli==1.32.36",
            "botocore==1.34.36",
            "rsa==4.7.2",
            "s3transfer==0.10.0",
        ],
        "dev": [
            "black==22.12.0",
            "isort>=5.10.1",
            "flake8>=4.0.1",
            "pytest>=7.4.4",
            "pre-commit==3.6.0",
        ],
    }

    extras_require.update({"all": [i[0] for i in extras_require.values()]})


setup(
    name="yvestest",
    version="0.1.1.dev0",
    packages=find_packages(),
    author="Dataverse Team",
    author_email="77601251@daum.net",
    description="An open-source simplifies ETL workflow with Python based on Spark",
    # url="github.com/UpstageAI/dataverse",
    license="MIT",
    include_package_data=True,
    install_requires=get_requirements(),
    extras_require=get_extras_require(),
    entry_points={"console_scripts": ["dataverse = dataverse.api.cli:main"]},
    python_requires=">=3.10, <=3.11",
)




"""
main entry point for the dataverse CLI tool
"""

from dataverse.utils.setting import SystemSetting


def main():
    """Main entry point for the cli."""
    print("🌌 Hello Welcome to Dataverse! 🌌")
    print("=" * 50)
    print("We are still under construction for CLI!")
    print("=" * 50)
    print("QUARK - By Ducky 🦆")

    # set the system setting to CLI mode
    SystemSetting().IS_CLI = True




from pyspark.rdd import RDD
from pyspark.sql import DataFrame
from pyspark.sql.functions import split, posexplode
from pyspark.sql.functions import collect_list

from dataverse.etl.registry import register_etl

import re
import unicodedata
from typing import Union



UNICODE_PUNCT = {
    "，": ",",
    "。": ".",
    "、": ",",
    "„": '"',
    "”": '"',
    "“": '"',
    "«": '"',
    "»": '"',
    "１": '"',
    "」": '"',
    "「": '"',
    "《": '"',
    "》": '"',
    "´": "'",
    "∶": ":",
    "：": ":",
    "？": "?",
    "！": "!",
    "（": "(",
    "）": ")",
    "；": ";",
    "–": "-",
    "—": " - ",
    "．": ". ",
    "～": "~",
    "’": "'",
    "…": "...",
    "━": "-",
    "〈": "<",
    "〉": ">",
    "【": "[",
    "】": "]",
    "％": "%",
    "►": "-",
}


@register_etl
def cleaning___unicode___remove_punct(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    *args,
    **kwargs
):
    """
    remove all the unicode punctuations

    Code is from facebookresearch/cc_net
    https://github.com/facebookresearch/cc_net/blob/main/cc_net/text_normalizer.py

    args:
        subset (str): subset or columns to consider
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    def _remove_unicode_punct(row):
        row[subset] = re.sub(f"[{''.join(UNICODE_PUNCT.keys())}]", "", row[subset])
        return row

    data = data.map(_remove_unicode_punct)

    return data


@register_etl
def cleaning___unicode___replace_punct(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    *args,
    **kwargs
):
    """
    replace all the unicode punctuations

    Code is from facebookresearch/cc_net
    https://github.com/facebookresearch/cc_net/blob/main/cc_net/text_normalizer.py

    args:
        subset (str): subset or columns to consider if duplicated
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    def _replace_unicode_punct(row):
        row[subset] = "".join((UNICODE_PUNCT.get(c, c) for c in row[subset]))
        return row

    data = data.map(_replace_unicode_punct)

    return data

@register_etl
def cleaning___unicode___normalize(
    spark,
    data: Union[RDD, DataFrame],
    subset='text',
    *args,
    **kwargs
):
    """
    normalize the unicode

    Args:
        subset (str): subset or columns to consider for normalization
    """
    if isinstance(data, DataFrame):
        data = data.rdd
        data = data.map(lambda row: row.asDict())

    def _normalize(row):
        row[subset] = unicodedata.normalize("NFC", row[subset])
        return row

    data = data.map(_normalize)
    return data
